# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib.api import validators
from neutron_lib.callbacks import events
from neutron_lib.callbacks import priority_group
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as const
from neutron_lib.db import api as db_api
from neutron_lib.exceptions import l3 as l3_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging

from neutron.db import l3_attrs_db
from neutron.objects import router
from neutron.services.l3_router.service_providers import base


LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class UserDefined(base.L3ServiceProvider):
    ha_support = base.OPTIONAL

    def __init__(self, l3_plugin):
        super().__init__(l3_plugin)
        self._user_defined_provider = __name__ + "." + self.__class__.__name__

    @property
    def _flavor_plugin(self):
        try:
            return self._flavor_plugin_ref
        except AttributeError:
            self._flavor_plugin_ref = directory.get_plugin(
                plugin_constants.FLAVORS)
            return self._flavor_plugin_ref

    def _is_user_defined_provider(self, context, router):
        flavor_id = router.get('flavor_id')
        if flavor_id is None or flavor_id is const.ATTR_NOT_SPECIFIED:
            return False
        flavor = self._flavor_plugin.get_flavor(context, flavor_id)
        provider = self._flavor_plugin.get_flavor_next_provider(
            context, flavor['id'])[0]
        return str(provider['driver']) == self._user_defined_provider

    @registry.receives(resources.ROUTER_CONTROLLER,
                       [events.PRECOMMIT_ADD_ASSOCIATION])
    def _process_router_add_association(self, resource, event, trigger,
                                        payload=None):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to associate user defined flavor to router %s',
                  router)

    def _is_ha(self, router):
        ha = router.get('ha')
        if not validators.is_attr_set(ha):
            ha = cfg.CONF.l3_ha
        return ha

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_CREATE],
                       priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
    def _process_precommit_router_create(self, resource, event, trigger,
                                         payload):
        router = payload.latest_state
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        router_db = payload.metadata['router_db']
        is_ha = self._is_ha(router)
        router['ha'] = is_ha
        l3_attrs_db.ExtraAttributesMixin.set_extra_attr_value(router_db,
                                                              'ha', is_ha)

    @registry.receives(resources.ROUTER, [events.AFTER_CREATE])
    def _process_router_create(self, resource, event, trigger, payload=None):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to create a user defined flavor router %s',
                  router)

    @registry.receives(resources.ROUTER, [events.AFTER_UPDATE])
    def _process_router_update(self, resource, event, trigger, payload=None):
        original = payload.states[0]
        updated = payload.states[1]
        context = payload.context
        if not self._is_user_defined_provider(context, original):
            # flavor_id attribute is not allowed in router PUTs, so we only
            # need to check the original router
            return
        router_id = payload.resource_id
        LOG.debug('Got request to update a user defined flavor router with id '
                  '%s. Original: %s. Updated: %s', router_id, original,
                  updated)

    @registry.receives(resources.ROUTER, [events.AFTER_DELETE])
    def _process_router_delete(self, resource, event, trigger, payload=None):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        router_id = payload.resource_id
        LOG.debug('Got request to delete a user defined flavor router with ',
                  'id %s:', router_id)

    @registry.receives(resources.ROUTER_INTERFACE, [events.AFTER_CREATE])
    def _process_add_router_interface(self, resource, event, trigger, payload):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        port = payload.metadata['port']
        subnets = payload.metadata['subnets']
        router_interface_info = self.l3plugin._make_router_interface_info(
            router.id, port['tenant_id'], port['id'], port['network_id'],
            subnets[-1]['id'], [subnet['id'] for subnet in subnets])
        LOG.debug('Got request to add interface %s to a user defined flavor '
                  'router with id %s', router_interface_info, router.id)

    @registry.receives(resources.ROUTER_INTERFACE, [events.AFTER_DELETE])
    def _process_remove_router_interface(self, resource, event, trigger,
                                         payload):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        subnet_ids = payload.metadata['subnet_ids']
        LOG.debug('Got request to remove interface to subnets %s from a user '
                  'defined flavor router with id %s', subnet_ids, router.id)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_CREATE])
    def _process_floatingip_create(self, resource, event, trigger, payload):
        context = payload.context
        fip = payload.states[0]
        if not fip['router_id']:
            return
        router = self.l3plugin.get_router(context, fip['router_id'])
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to create a floating ip associated to a router '
                  'of user defined flavor %s', fip)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_UPDATE])
    def _process_floatingip_update(self, resource, event, trigger, payload):
        context = payload.context
        fip = payload.states[1]
        if not fip['router_id']:
            return
        router = self.l3plugin.get_router(context, fip['router_id'])
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to update a floating ip associated to a router '
                  'of user defined flavor %s', fip)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_DELETE])
    def _process_floatingip_delete(self, resource, event, trigger, payload):
        context = payload.context
        fip = payload.states[0]
        if not fip['router_id']:
            return
        router = self.l3plugin.get_router(context, fip['router_id'])
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to delete a floating ip associated to a router '
                  'of user defined flavor %s', fip)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_STATUS_UPDATE])
    def _process_floatingip_status_update(self, resource, event, trigger,
                                          payload):
        context = payload.context
        fip = payload.states[0]
        if not fip['router_id']:
            return
        router = self.l3plugin.get_router(context, fip['router_id'])
        if not self._is_user_defined_provider(context, router):
            return
        LOG.debug('Got request to update the status of a floating ip '
                  'associated to a router of user defined flavor %s', fip)


@registry.has_registry_receivers
class UserDefinedNoLsp(UserDefined):

    @registry.receives(resources.ROUTER_INTERFACE, [events.AFTER_CREATE])
    def _process_add_router_interface(self, resource, event, trigger, payload):
        router = payload.states[0]
        context = payload.context
        if not self._is_user_defined_provider(context, router):
            return
        port = payload.metadata['port']
        subnets = payload.metadata['subnets']
        router_interface_info = self.l3plugin._make_router_interface_info(
            router.id, port['tenant_id'], port['id'], port['network_id'],
            subnets[-1]['id'], [subnet['id'] for subnet in subnets])
        self.l3plugin._ovn_client.delete_port(context, port['id'],
                                              port_object=port)
        LOG.debug('Got request to add interface %s to a user defined flavor '
                  'router with id %s. The OVN LSP was deleted.',
                  router_interface_info, router.id)

    def _get_port_to_recreate(self, context, router_id, subnet_id):
        with db_api.CONTEXT_READER.using(context):
            objs = router.RouterPort.get_objects(
                context, router_id=router_id,
                port_type=const.DEVICE_OWNER_ROUTER_INTF)
            router_ports = [
                self.l3plugin._plugin._make_port_dict(rp.db_obj.port)
                for rp in objs]
        for rp in router_ports:
            for ip in rp['fixed_ips']:
                if ip['subnet_id'] == subnet_id:
                    return rp
        raise l3_exc.RouterInterfaceNotFoundForSubnet(router_ports=router_id,
                                                      subnet_id=subnet_id)

    @registry.receives(resources.ROUTER_INTERFACE, [events.BEFORE_DELETE])
    def _process_before_remove_router_interface(self, resource, event, trigger,
                                                payload):
        context = payload.context
        router_id = payload.resource_id
        router_obj = router.Router.get_object(context, id=router_id)
        if not self._is_user_defined_provider(context, router_obj):
            return
        subnet_id = payload.metadata['subnet_id']
        port = self._get_port_to_recreate(context, router_id, subnet_id)

        # If the LSP exists, the interface is being removed by port and the
        # port has fixed ips in more than one subnet, then the port has been
        # recreated in a previous notification of this event. Do nothing
        nbdb_idl = self.l3plugin._ovn_client._nb_idl
        if nbdb_idl.lookup('Logical_Switch_Port', port['id'], default=None):
            return

        # The ovn client creates the revision row when creating the LSP
        self.l3plugin._ovn_client.create_port(context, port)
        LOG.debug('Recreated OVN LSP for port with id %s before removing '
                  'interface for user defined flavor router with id %s',
                  port['id'], router_id)



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils

from neutron import privileged

COMMAND = 'ethtool'


@privileged.default.entrypoint
def offload(device, rx, tx, namespace=None):
    cmd = []
    if namespace:
        cmd += ['ip', 'netns', 'exec', namespace]
    rx = 'on' if rx else 'off'
    tx = 'on' if tx else 'off'
    cmd += [COMMAND, '--offload', device, 'rx', rx, 'tx', tx]
    return processutils.execute(*cmd)



# Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants
from neutron_lib.services.qos import constants as qos_consts
from neutronclient.common import exceptions
from oslo_utils import uuidutils

from neutron.common import utils
from neutron.tests.common.agents import l2_extensions
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import config as fullstack_config
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine
from neutron.tests.unit import testlib_api

from neutron.agent.common import ovs_lib
from neutron.plugins.ml2.common import constants as comm_consts


load_tests = testlib_api.module_load_tests

BANDWIDTH_BURST = 100
BANDWIDTH_LIMIT = 500
MIN_BANDWIDTH = 300
DSCP_MARK = 16

PACKET_RATE_LIMIT = 10000
PACKET_RATE_BURST = 1000


class BaseQoSRuleTestCase:
    number_of_hosts = 1
    physical_network = None

    @property
    def reverse_direction(self):
        if self.direction == constants.INGRESS_DIRECTION:
            return constants.EGRESS_DIRECTION
        if self.direction == constants.EGRESS_DIRECTION:
            return constants.INGRESS_DIRECTION

    def setUp(self):
        host_desc = [
            environment.HostDescription(
                l3_agent=False,
                l2_agent_type=self.l2_agent_type
            ) for _ in range(self.number_of_hosts)]
        env_desc = environment.EnvironmentDescription(
            agent_down_time=10,
            qos=True)
        env = environment.Environment(env_desc, host_desc)
        super().setUp(env)
        self.l2_agent_process = self.environment.hosts[0].l2_agent
        self.l2_agent = self.safe_client.client.list_agents(
            agent_type=self.l2_agent_type)['agents'][0]

        self.tenant_id = uuidutils.generate_uuid()
        network_args = {}
        if self.physical_network:
            network_args = {'physical_network': self.physical_network,
                            'network_type': 'vlan'}
        self.network = self.safe_client.create_network(
            self.tenant_id, name='network-test', **network_args)
        self.subnet = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            name='subnet-test',
            enable_dhcp=False)

    def _create_qos_policy(self):
        return self.safe_client.create_qos_policy(
            self.tenant_id, 'fs_policy', 'Fullstack testing policy',
            shared='False', is_default='False')

    def _prepare_vm_with_qos_policy(self, rule_add_functions):
        if rule_add_functions:
            qos_policy = self._create_qos_policy()
            qos_policy_id = qos_policy['id']
            for rule_add in rule_add_functions:
                rule_add(qos_policy)
        else:
            qos_policy_id = qos_policy = None

        port = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            qos_policy_id)

        vm = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port))

        return vm, qos_policy


class _TestBwLimitQoS(BaseQoSRuleTestCase):

    number_of_hosts = 1

    @staticmethod
    def _get_expected_egress_burst_value(limit):
        return int(
            limit * qos_consts.DEFAULT_BURST_RATE
        )

    def _wait_for_bw_rule_removed(self, vm, direction):
        # No values are provided when port doesn't have qos policy
        self._wait_for_bw_rule_applied(vm, None, None, direction)

    def _add_bw_limit_rule(self, limit, burst, direction, qos_policy):
        qos_policy_id = qos_policy['id']
        rule = self.safe_client.create_bandwidth_limit_rule(
            qos_policy_id, limit, burst, direction)
        # Make it consistent with GET reply
        rule['type'] = qos_consts.RULE_TYPE_BANDWIDTH_LIMIT
        rule['qos_policy_id'] = qos_policy_id
        qos_policy['rules'].append(rule)

    def _create_vm_with_limit_rules(self):
        # Create port with qos policy attached, with different direction
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_bw_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction),
             functools.partial(
                self._add_bw_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.reverse_direction)])

        self._wait_for_bw_rule_applied(
            vm, BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)
        self._wait_for_bw_rule_applied(
            vm, BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.reverse_direction)
        return vm, qos_policy

    def _restart_agent_and_check_rules_applied(self, policy_id, vm,
                                               final_rules,
                                               add_rules=None,
                                               update_rules=None,
                                               delete_rules=None):
        # final_rules: the last valid rule after all operations
        # (clear/update/reset rules during the l2-agent stop) are completed.
        # add_rules: rules that need to be added during the l2-agent stop.
        # update_rules: rules that need to be updated during the l2-agent stop.
        # delete_rules:rules that need to be deleted during the l2-agent stop.

        add_rules = [] if not add_rules else add_rules
        update_rules = [] if not update_rules else update_rules
        delete_rules = [] if not delete_rules else delete_rules
        # Stop l2_agent and clear/update/reset the port qos rules
        self.l2_agent_process.stop()
        self._wait_until_agent_down(self.l2_agent['id'])

        for rule in delete_rules:
            self.client.delete_bandwidth_limit_rule(rule['id'],
                                                    policy_id)

        for rule in add_rules:
            self.safe_client.create_bandwidth_limit_rule(
                policy_id, rule.get('limit'), rule.get('burst'),
                rule['direction'])

        for rule in update_rules:
            self.client.update_bandwidth_limit_rule(
                rule['id'], policy_id,
                body={'bandwidth_limit_rule':
                      {'max_kbps': rule.get('limit'),
                       'max_burst_kbps': rule.get('burst'),
                       'direction': rule.get('direction')}})

        # Start l2_agent to check if these rules is cleared
        self.l2_agent_process.start()
        self._wait_until_agent_up(self.l2_agent['id'])

        all_directions = {self.direction, self.reverse_direction}
        for final_rule in final_rules:
            all_directions -= {final_rule['direction']}
            self._wait_for_bw_rule_applied(
                vm, final_rule.get('limit'),
                final_rule.get('burst'), final_rule['direction'])
        # Make sure there are no other rules.
        for direction in list(all_directions):
            self._wait_for_bw_rule_applied(vm, None, None, direction)

    def test_bw_limit_direction_change(self):
        # Create port with qos policy attached, with rule self.direction
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_bw_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)])
        bw_rule = qos_policy['rules'][0]

        self._wait_for_bw_rule_applied(
            vm, BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)

        # Update rule by changing direction to opposite then it was before
        self.client.update_bandwidth_limit_rule(
            bw_rule['id'], qos_policy['id'],
            body={'bandwidth_limit_rule': {
                'direction': self.reverse_direction}})
        self._wait_for_bw_rule_removed(vm, self.direction)
        self._wait_for_bw_rule_applied(
            vm, BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.reverse_direction)

    def test_bw_limit_qos_no_rules_l2_agent_restart(self):
        vm, qos_policy = self._create_vm_with_limit_rules()

        bw_rule_1 = qos_policy['rules'][0]
        bw_rule_2 = qos_policy['rules'][1]
        qos_policy_id = qos_policy['id']

        # final_rules indicates the last valid rule after all operations
        # (clear/update/reset rules during the l2-agent stop) are completed
        final_rules = [{'direction': self.direction,
                        'limit': None},
                       {'direction': self.reverse_direction,
                        'limit': None}]

        self._restart_agent_and_check_rules_applied(
            qos_policy_id, vm, final_rules=final_rules,
            delete_rules=[bw_rule_1, bw_rule_2])

    def test_bw_limit_qos_rules_deleted_l2_agent_restart(self):
        vm, qos_policy = self._create_vm_with_limit_rules()

        bw_rule_1 = qos_policy['rules'][0]
        qos_policy_id = qos_policy['id']

        # final_rules indicates the last valid rule after all operations
        # (clear/update/reset rules during the l2-agent stop) are completed
        final_rules = [{'direction': self.direction,
                        'limit': None},
                       {'direction': self.reverse_direction,
                        'limit': BANDWIDTH_LIMIT,
                        'burst': BANDWIDTH_BURST}]

        self._restart_agent_and_check_rules_applied(
            qos_policy_id, vm, final_rules=final_rules,
            delete_rules=[bw_rule_1])

    def test_bw_limit_qos_rules_changed_l2_agent_restart(self):
        vm, qos_policy = self._create_vm_with_limit_rules()

        bw_rule_1 = qos_policy['rules'][0]
        bw_rule_2 = qos_policy['rules'][1]
        qos_policy_id = qos_policy['id']

        add_rules = [{'direction': self.direction,
                      'limit': BANDWIDTH_LIMIT * 2,
                      'burst': BANDWIDTH_BURST * 2},
                     {'direction': self.reverse_direction,
                      'limit': BANDWIDTH_LIMIT * 2,
                      'burst': BANDWIDTH_BURST * 2}]

        self._restart_agent_and_check_rules_applied(
            qos_policy_id, vm, final_rules=add_rules,
            add_rules=add_rules,
            delete_rules=[bw_rule_1, bw_rule_2])

    def test_bw_limit_qos_rules_updated_l2_agent_restart(self):
        vm, qos_policy = self._create_vm_with_limit_rules()

        bw_rule_1 = qos_policy['rules'][0]
        bw_rule_2 = qos_policy['rules'][1]
        qos_policy_id = qos_policy['id']

        update_rules = [{'id': bw_rule_1['id'],
                         'direction': bw_rule_1['direction'],
                         'limit': BANDWIDTH_LIMIT * 2,
                         'burst': BANDWIDTH_BURST * 2},
                        {'id': bw_rule_2['id'],
                         'direction': bw_rule_2['direction'],
                         'limit': BANDWIDTH_LIMIT * 2,
                         'burst': BANDWIDTH_BURST * 2}]

        self._restart_agent_and_check_rules_applied(
            qos_policy_id, vm, final_rules=update_rules,
            update_rules=update_rules)


class TestBwLimitQoSOvs(_TestBwLimitQoS, base.BaseFullStackTestCase):
    l2_agent_type = constants.AGENT_TYPE_OVS
    scenarios = [
        ('ingress', {'direction': constants.INGRESS_DIRECTION}),
        ('egress', {'direction': constants.EGRESS_DIRECTION})
    ]

    @staticmethod
    def _get_expected_burst_value(limit, direction):
        if direction != constants.EGRESS_DIRECTION:
            return 0
        # For egress bandwidth limit this value should be calculated as
        # bandwidth_limit * qos_consts.DEFAULT_BURST_RATE
        return TestBwLimitQoSOvs._get_expected_egress_burst_value(limit)

    def _wait_for_bw_rule_applied(self, vm, limit, burst, direction):
        if direction == constants.EGRESS_DIRECTION:
            base.wait_until_true(
                lambda: vm.bridge.get_egress_bw_limit_for_port(
                    vm.port.name) == (limit, burst))
        elif direction == constants.INGRESS_DIRECTION:
            base.wait_until_true(
                lambda: vm.bridge.get_ingress_bw_limit_for_port(
                    vm.port.name) == (limit, burst),
                timeout=10)
            br_int_flows = vm.bridge.dump_flows_for_table(
                ovs_constants.LOCAL_SWITCHING)
            expected = (
                'priority=200,reg3=0 '
                'actions=set_queue:%(queue_num)s,'
                'load:0x1->NXM_NX_REG3[0],resubmit(,0)' % {
                    'queue_num': ovs_lib.QOS_DEFAULT_QUEUE
                }
            )
            self.assertIn(expected, br_int_flows)

    def test_bw_limit_qos_port_removed(self):
        """Test if rate limit config is properly removed when whole port is
        removed.
        """

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_bw_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)])
        self._wait_for_bw_rule_applied(
            vm, BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)

        # Delete port with qos policy attached
        vm.destroy(delete_port=True)
        self._wait_for_bw_rule_removed(vm, self.direction)
        qos_id, qos_queues = vm.bridge._find_qos(
            vm.port.name,
            qos_consts.RULE_TYPE_BANDWIDTH_LIMIT)
        self.assertIsNone(qos_id)
        self.assertIsNone(qos_queues)


class _TestDscpMarkingQoS(BaseQoSRuleTestCase):

    number_of_hosts = 2

    def _wait_for_dscp_marking_rule_removed(self, vm):
        self._wait_for_dscp_marking_rule_applied(vm, None)

    def _add_dscp_rule(self, dscp_mark, qos_policy):
        qos_policy_id = qos_policy['id']
        rule = self.safe_client.create_dscp_marking_rule(
            qos_policy_id, dscp_mark)
        # Make it consistent with GET reply
        rule['type'] = qos_consts.RULE_TYPE_DSCP_MARKING
        rule['qos_policy_id'] = qos_policy_id
        qos_policy['rules'].append(rule)

    def test_dscp_qos_policy_rule_lifecycle(self):
        new_dscp_mark = DSCP_MARK + 8

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(self._add_dscp_rule, DSCP_MARK)])
        dscp_rule = qos_policy['rules'][0]

        self._wait_for_dscp_marking_rule_applied(vm, DSCP_MARK)
        qos_policy_id = qos_policy['id']

        self.client.delete_dscp_marking_rule(dscp_rule['id'], qos_policy_id)
        self._wait_for_dscp_marking_rule_removed(vm)

        # Create new rule
        new_rule = self.safe_client.create_dscp_marking_rule(
            qos_policy_id, new_dscp_mark)
        self._wait_for_dscp_marking_rule_applied(vm, new_dscp_mark)

        # Update qos policy rule id
        self.client.update_dscp_marking_rule(
            new_rule['id'], qos_policy_id,
            body={'dscp_marking_rule': {'dscp_mark': DSCP_MARK}})
        self._wait_for_dscp_marking_rule_applied(vm, DSCP_MARK)

        # Remove qos policy from port
        self.client.update_port(
            vm.neutron_port['id'],
            body={'port': {'qos_policy_id': None}})
        self._wait_for_dscp_marking_rule_removed(vm)

    def test_dscp_marking_packets(self):
        # Create port (vm) which will be used to received and test packets
        receiver_port = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[1].hostname)

        receiver = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[1],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=receiver_port))

        # Create port with qos policy attached
        sender, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(self._add_dscp_rule, DSCP_MARK)])

        sender.block_until_boot()
        receiver.block_until_boot()

        self._wait_for_dscp_marking_rule_applied(sender, DSCP_MARK)
        l2_extensions.wait_for_dscp_marked_packet(
            sender, receiver, DSCP_MARK)

    def test_dscp_marking_clean_port_removed(self):
        """Test if DSCP marking OpenFlow/iptables rules are removed when
        whole port is removed.
        """

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(self._add_dscp_rule, DSCP_MARK)])

        self._wait_for_dscp_marking_rule_applied(vm, DSCP_MARK)

        # Delete port with qos policy attached
        vm.destroy(delete_port=True)
        self._wait_for_dscp_marking_rule_removed(vm)


class TestDscpMarkingQoSOvs(_TestDscpMarkingQoS, base.BaseFullStackTestCase):
    l2_agent_type = constants.AGENT_TYPE_OVS

    def _wait_for_dscp_marking_rule_applied(self, vm, dscp_mark):
        l2_extensions.wait_until_dscp_marking_rule_applied_ovs(
            vm.bridge, vm.port.name, dscp_mark)


class _TestMeterRateLimitQoS(object):

    def _wait_for_meter_rate_limit_rule_applied(
            self, vm, direction,
            type_=comm_consts.METER_FLAG_PPS):
        l2_extensions.wait_until_pkt_meter_rule_applied_ovs(
            vm.bridge, vm.port.name, vm.neutron_port['id'],
            direction, vm.mac_address, type_=type_)

    def _wait_for_meter_rate_limit_rule_removed(
            self, vm, direction,
            type_=comm_consts.METER_FLAG_PPS):
        l2_extensions.wait_until_pkt_meter_rule_applied_ovs(
            vm.bridge, vm.port.name, vm.neutron_port['id'], direction,
            type_=type_)

    def _add_packet_rate_limit_rule(self, limit, burst, direction, qos_policy):
        qos_policy_id = qos_policy['id']
        rule = self.safe_client.create_packet_rate_limit_rule(
            qos_policy_id, limit, burst, direction)
        rule['type'] = qos_consts.RULE_TYPE_PACKET_RATE_LIMIT
        rule['qos_policy_id'] = qos_policy_id
        qos_policy['rules'].append(rule)

    def _add_bandwidth_rate_limit_rule(self, limit, burst,
                                       direction, qos_policy):
        qos_policy_id = qos_policy['id']
        rule = self.safe_client.create_bandwidth_limit_rule(
            qos_policy_id, limit, burst, direction)
        rule['type'] = qos_consts.RULE_TYPE_BANDWIDTH_LIMIT
        rule['qos_policy_id'] = qos_policy_id
        qos_policy['rules'].append(rule)

    def _create_vm_with_limit_rules(self):
        # Create port with qos policy attached, with different direction
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_packet_rate_limit_rule,
                PACKET_RATE_LIMIT, PACKET_RATE_BURST, self.direction),
             functools.partial(
                self._add_packet_rate_limit_rule,
                PACKET_RATE_LIMIT, PACKET_RATE_BURST, self.reverse_direction),
             functools.partial(
                self._add_bandwidth_rate_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction),
             functools.partial(
                self._add_bandwidth_rate_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.reverse_direction)])

        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction)
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.reverse_direction)
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction, type_=comm_consts.METER_FLAG_BPS)
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.reverse_direction, type_=comm_consts.METER_FLAG_BPS)
        return vm, qos_policy

    def test_packet_rate_limit_qos_policy_rule_lifecycle(self):
        new_limit = PACKET_RATE_LIMIT + 100

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_packet_rate_limit_rule,
                PACKET_RATE_LIMIT, PACKET_RATE_BURST, self.direction)])

        vm.bridge.use_at_least_protocol(ovs_constants.OPENFLOW13)
        if not vm.bridge.list_meter_features():
            self.skip("Test ovs bridge %s does not support meter.",
                      vm.bridge.br_name)

        pkt_rule = qos_policy['rules'][0]
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction)
        qos_policy_id = qos_policy['id']

        self.client.delete_packet_rate_limit_rule(pkt_rule['id'],
                                                  qos_policy_id)
        self._wait_for_meter_rate_limit_rule_removed(vm, self.direction)

        new_rule = self.safe_client.create_packet_rate_limit_rule(
            qos_policy_id, new_limit, direction=self.direction)
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction)

        # Update qos policy rule id
        self.client.update_packet_rate_limit_rule(
            new_rule['id'], qos_policy_id,
            body={'packet_rate_limit_rule': {
                'max_kpps': PACKET_RATE_LIMIT,
                'max_burst_kpps': PACKET_RATE_BURST}})
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction)

        # Remove qos policy from port
        self.client.update_port(
            vm.neutron_port['id'],
            body={'port': {'qos_policy_id': None}})
        self._wait_for_meter_rate_limit_rule_removed(vm, self.direction)

    def test_meter_bandwidth_rate_limit_qos_policy_rule_lifecycle(self):
        new_limit = BANDWIDTH_LIMIT + 100

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_bandwidth_rate_limit_rule,
                BANDWIDTH_LIMIT, BANDWIDTH_BURST, self.direction)])

        vm.bridge.use_at_least_protocol(ovs_constants.OPENFLOW13)
        if not vm.bridge.list_meter_features():
            self.skip("Test ovs bridge %s does not support meter.",
                      vm.bridge.br_name)

        pkt_rule = qos_policy['rules'][0]
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction, type_=comm_consts.METER_FLAG_BPS)
        qos_policy_id = qos_policy['id']

        self.client.delete_bandwidth_limit_rule(pkt_rule['id'],
                                                qos_policy_id)
        self._wait_for_meter_rate_limit_rule_removed(
            vm, self.direction,
            type_=comm_consts.METER_FLAG_BPS)

        new_rule = self.safe_client.create_bandwidth_limit_rule(
            qos_policy_id, new_limit, direction=self.direction)
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction, type_=comm_consts.METER_FLAG_BPS)

        # Update qos policy rule id
        self.client.update_bandwidth_limit_rule(
            new_rule['id'], qos_policy_id,
            body={'bandwidth_limit_rule': {
                'max_kbps': BANDWIDTH_LIMIT,
                'max_burst_kbps': BANDWIDTH_BURST}})
        self._wait_for_meter_rate_limit_rule_applied(
            vm, self.direction, type_=comm_consts.METER_FLAG_BPS)

        # Remove qos policy from port
        self.client.update_port(
            vm.neutron_port['id'],
            body={'port': {'qos_policy_id': None}})
        self._wait_for_meter_rate_limit_rule_removed(
            vm, self.direction, type_=comm_consts.METER_FLAG_BPS)


class TestMeterRateLimitQoSOvs(_TestMeterRateLimitQoS,
                               base.BaseFullStackTestCase):
    number_of_hosts = 1
    physical_network = None
    l2_agent_type = constants.AGENT_TYPE_OVS
    scenarios = [
        ('ingress', {'direction': constants.INGRESS_DIRECTION}),
        ('egress', {'direction': constants.EGRESS_DIRECTION})
    ]

    @property
    def reverse_direction(self):
        if self.direction == constants.INGRESS_DIRECTION:
            return constants.EGRESS_DIRECTION
        elif self.direction == constants.EGRESS_DIRECTION:
            return constants.INGRESS_DIRECTION

    def setUp(self):
        host_desc = [
            environment.HostDescription(
                l3_agent=False,
                l2_agent_type=self.l2_agent_type
            ) for _ in range(self.number_of_hosts)]
        env_desc = environment.EnvironmentDescription(
            agent_down_time=10,
            qos=True,
            use_meter_bandwidth_limit=True)
        env = environment.Environment(env_desc, host_desc)
        super(TestMeterRateLimitQoSOvs, self).setUp(env)
        self.l2_agent_process = self.environment.hosts[0].l2_agent
        self.l2_agent = self.safe_client.client.list_agents(
            agent_type=self.l2_agent_type)['agents'][0]

        self.tenant_id = uuidutils.generate_uuid()
        network_args = {}
        if self.physical_network:
            network_args = {'physical_network': self.physical_network,
                            'network_type': 'vlan'}
        self.network = self.safe_client.create_network(
            self.tenant_id, name='network-test', **network_args)
        self.subnet = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            name='subnet-test',
            enable_dhcp=False)

    def _create_qos_policy(self):
        return self.safe_client.create_qos_policy(
            self.tenant_id, 'fs_policy', 'Fullstack testing policy',
            shared='False', is_default='False')

    def _prepare_vm_with_qos_policy(self, rule_add_functions):
        if rule_add_functions:
            qos_policy = self._create_qos_policy()
            qos_policy_id = qos_policy['id']
            for rule_add in rule_add_functions:
                rule_add(qos_policy)
        else:
            qos_policy_id = qos_policy = None

        port = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            qos_policy_id)

        vm = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port))

        return vm, qos_policy


class TestQoSPolicyIsDefault(base.BaseFullStackTestCase):

    NAME = 'fs_policy'
    DESCRIPTION = 'Fullstack testing policy'
    SHARED = True

    def setUp(self):
        host_desc = []  # No need to register agents for this test case
        env_desc = environment.EnvironmentDescription(qos=True)
        env = environment.Environment(env_desc, host_desc)
        super().setUp(env)

    def _create_qos_policy(self, project_id, is_default):
        return self.safe_client.create_qos_policy(
            project_id, self.NAME, self.DESCRIPTION, shared=self.SHARED,
            is_default=is_default)

    def _update_qos_policy(self, qos_policy_id, is_default):
        return self.client.update_qos_policy(
            qos_policy_id, body={'policy': {'is_default': is_default}})

    def test_qos_policy_is_default(self):
        # 1) Create one default QoS policy per project.
        project_ids = [uuidutils.generate_uuid(), uuidutils.generate_uuid()]
        for project_id in project_ids:
            qos_policy1 = self._create_qos_policy(project_id, True)
            self.assertTrue(qos_policy1['is_default'])
            self.assertEqual(project_id, qos_policy1['project_id'])
            qos_policy2 = self._create_qos_policy(project_id, False)
            self.assertFalse(qos_policy2['is_default'])
            self.assertEqual(project_id, qos_policy2['project_id'])

        # 2) Try to add a second default policy to the first project, that has
        # already one QoS default policy
        self.assertRaises(exceptions.Conflict,
                          self._create_qos_policy, project_ids[0], True)

        # 3) Update default status of a QoS policy.
        qos_policy1 = self._update_qos_policy(qos_policy1['id'], False)
        self.assertFalse(qos_policy1['policy']['is_default'])
        qos_policy2 = self._update_qos_policy(qos_policy2['id'], True)
        self.assertTrue(qos_policy2['policy']['is_default'])

        # 4) Try to update the QoS policy status if a default QoS policy
        # already exists in this project.
        qos_policy_3 = self._create_qos_policy(project_ids[1], False)
        self.assertFalse(qos_policy_3['is_default'])
        self.assertRaises(exceptions.Conflict,
                          self._update_qos_policy, qos_policy_3['id'], True)


class _TestMinBwQoS(BaseQoSRuleTestCase):

    number_of_hosts = 1
    physical_network = fullstack_config.PHYSICAL_NETWORK_NAME

    def _wait_for_min_bw_rule_removed(self, vm, direction):
        # No values are provided when port doesn't have qos policy
        self._wait_for_min_bw_rule_applied(vm, None, direction)

    def _add_min_bw_rule(self, min_bw, direction, qos_policy):
        qos_policy_id = qos_policy['id']
        rule = self.safe_client.create_minimum_bandwidth_rule(
            qos_policy_id, min_bw, direction)
        # Make it consistent with GET reply
        rule['type'] = qos_consts.RULE_TYPE_MINIMUM_BANDWIDTH
        rule['qos_policy_id'] = qos_policy_id
        qos_policy['rules'].append(rule)

    def test_min_bw_qos_policy_rule_lifecycle(self):
        new_limit = MIN_BANDWIDTH - 100

        # Create port with qos policy attached
        vm, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_min_bw_rule, MIN_BANDWIDTH, self.direction)])
        bw_rule = qos_policy['rules'][0]

        self._wait_for_min_bw_rule_applied(vm, MIN_BANDWIDTH, self.direction)
        qos_policy_id = qos_policy['id']

        self.client.delete_minimum_bandwidth_rule(bw_rule['id'], qos_policy_id)
        self._wait_for_min_bw_rule_removed(vm, self.direction)

        new_rule = self.safe_client.create_minimum_bandwidth_rule(
            qos_policy_id, new_limit, direction=self.direction)
        self._wait_for_min_bw_rule_applied(vm, new_limit, self.direction)

        # Update qos policy rule id
        self.client.update_minimum_bandwidth_rule(
            new_rule['id'], qos_policy_id,
            body={'minimum_bandwidth_rule': {'min_kbps': MIN_BANDWIDTH}})
        self._wait_for_min_bw_rule_applied(vm, MIN_BANDWIDTH, self.direction)

        # Remove qos policy from port
        self.client.update_port(
            vm.neutron_port['id'],
            body={'port': {'qos_policy_id': None}})
        self._wait_for_min_bw_rule_removed(vm, self.direction)


class TestMinBwQoSOvs(_TestMinBwQoS, base.BaseFullStackTestCase):
    l2_agent_type = constants.AGENT_TYPE_OVS
    scenarios = [
        ('egress', {'direction': constants.EGRESS_DIRECTION})
    ]

    def _wait_for_min_bw_rule_applied(self, vm, min_bw, direction):
        if direction == constants.EGRESS_DIRECTION:
            try:
                base.wait_until_true(
                    lambda: vm.bridge.get_egress_min_bw_for_port(
                        vm.neutron_port['id']) == min_bw)
            except utils.WaitTimeout:
                qoses, queues = self._qos_info(vm.bridge)
                msg = ('QoS for port %s failed to apply minimum BW rule %s' %
                       (vm.neutron_port['id'], min_bw))
                self.fail(msg + qoses + queues)

        elif direction == constants.INGRESS_DIRECTION:
            self.fail('"%s" direction not implemented'
                      % constants.INGRESS_DIRECTION)

    def _qos_info(self, vm_bridge):
        qoses = vm_bridge._list_qos()
        queues = vm_bridge._list_queues()
        qoses = '\nList of OVS QoS registers:\n%s' % ''.join(str(qoses))
        queues = '\nList of OVS Queue registers:\n%s' % ''.join(str(queues))
        return qoses, queues

    def _find_agent_qos_and_queue(self, vm):
        # NOTE(ralonsoh): the "_min_bw_qos_id" in vm.bridge is not the same as
        # the ID in the agent br_int instance. We need first to find the QoS
        # register and the Queue assigned to vm.neutron_port['id']
        data = {'qos': None, 'qos_queue': None, 'queue_num': None}

        def check_qos_and_queue():
            queue = vm.bridge._find_queue(vm.neutron_port['id'])
            data['queue_num'] = int(queue['external_ids']['queue-num'])
            qoses = vm.bridge._list_qos()
            for qos in qoses:
                qos_queue = qos['queues'].get(data['queue_num'])
                if qos_queue and qos_queue.uuid == queue['_uuid']:
                    data['qos'] = qos
                    data['qos_queue'] = qos_queue
                    return True

        try:
            base.wait_until_true(check_qos_and_queue, timeout=10)
            return data['qos'], data['qos_queue']
        except utils.WaitTimeout:
            queuenum = ('QoS register not found with queue-num %s' %
                        data['queue_num'])
            qoses, queues = self._qos_info(vm.bridge)
            self.fail(queuenum + qoses + queues)

    def test_min_bw_qos_create_network_vxlan_supported(self):
        qos_policy = self._create_qos_policy()
        qos_policy_id = qos_policy['id']
        self.safe_client.create_minimum_bandwidth_rule(
            qos_policy_id, MIN_BANDWIDTH, self.direction)
        network_args = {'network_type': 'vxlan',
                        'qos_policy_id': qos_policy_id}
        net = self.safe_client.create_network(
            self.tenant_id, name='network-test', **network_args)
        self.assertEqual(qos_policy_id, net['qos_policy_id'])

    def test_min_bw_qos_create_and_update_network_vxlan_supported(self):
        qos_policy = self._create_qos_policy()
        qos_policy_id = qos_policy['id']
        self.safe_client.create_minimum_bandwidth_rule(
            qos_policy_id, MIN_BANDWIDTH, self.direction)
        network_args = {'network_type': 'vxlan',
                        'qos_policy_id': qos_policy_id}
        network = self.safe_client.create_network(
            self.tenant_id, name='network-test', **network_args)
        self.assertEqual(qos_policy_id, network['qos_policy_id'])

        qos_policy2 = self._create_qos_policy()
        qos_policy2_id = qos_policy2['id']
        self.client.update_network(
            network['id'], body={'network': {'qos_policy_id': qos_policy2_id}})
        _net = self.client.show_network(network['id'])
        self.assertEqual(qos_policy2_id, _net['network']['qos_policy_id'])

        # This action will remove the QoS policy from the network. This is also
        # necessary before the cleanUp call, that will delete the QoS policy
        # before the network.
        self.client.update_network(
            network['id'], body={'network': {'qos_policy_id': None}})
        _net = self.client.show_network(network['id'])
        self.assertIsNone(_net['network']['qos_policy_id'])

    def test_min_bw_qos_port_removed(self):
        """Test if min BW limit config is properly removed when port removed.

        In case another port is added without a QoS policy, the L2 agent QoS
        extension will call "handle_port" and then it will force the reset of
        this port (self._process_reset_port(port)). This test will check that
        if the port is not present in the agent QoS cache, the policy is not
        removed.
        """
        # Create port without qos policy attached
        vm_noqos, _ = self._prepare_vm_with_qos_policy(None)

        # Create port with qos policy attached
        vm_qos, qos_policy = self._prepare_vm_with_qos_policy(
            [functools.partial(
                self._add_min_bw_rule, MIN_BANDWIDTH, self.direction)])
        self._wait_for_min_bw_rule_applied(
            vm_qos, MIN_BANDWIDTH, self.direction)

        # Check QoS policy and Queue rule.
        qos, queue = self._find_agent_qos_and_queue(vm_qos)
        self.assertEqual({'min-rate': str(MIN_BANDWIDTH * 1000),
                          'max-rate': str(ovs_lib.OVS_MAX_RATE)},
                         queue.other_config)
        queues = vm_qos.bridge._list_queues(port=vm_qos.neutron_port['id'])
        self.assertEqual(1, len(queues))
        self.assertEqual(queue.uuid, queues[0]['_uuid'])

        # Delete port with qos policy attached
        vm_qos.destroy(delete_port=True)
        self._wait_for_min_bw_rule_removed(vm_qos, self.direction)
        self.assertEqual(
            [],
            vm_qos.bridge._list_queues(port=vm_qos.neutron_port['id']))

        vm_noqos.destroy()



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base


COLLECTION_PATH = '/agents'
RESOURCE_PATH = '/agents/{id}'

DEPRECATION_REASON = (
    "The Agent API now supports project scope and default roles.")

rules = [
    policy.DocumentedRuleDefault(
        name='create_agent',
        check_str=base.ADMIN,
        description='Create an agent',
        operations=[
            {
                'method': 'POST',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        name='get_agent',
        check_str=base.ADMIN,
        description='Get an agent',
        operations=[
            {
                'method': 'GET',
                'path': COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_agent',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_agent',
        check_str=base.ADMIN,
        description='Update an agent',
        operations=[
            {
                'method': 'PUT',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='update_agent',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_agent',
        check_str=base.ADMIN,
        description='Delete an agent',
        operations=[
            {
                'method': 'DELETE',
                'path': RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_agent',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_dhcp-network',
        check_str=base.ADMIN,
        description='Add a network to a DHCP agent',
        operations=[
            {
                'method': 'POST',
                'path': '/agents/{agent_id}/dhcp-networks',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='create_dhcp-network',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_dhcp-networks',
        check_str=base.ADMIN,
        description='List networks on a DHCP agent',
        operations=[
            {
                'method': 'GET',
                'path': '/agents/{agent_id}/dhcp-networks',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_dhcp-networks',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_dhcp-network',
        check_str=base.ADMIN,
        description='Remove a network from a DHCP agent',
        operations=[
            {
                'method': 'DELETE',
                'path': '/agents/{agent_id}/dhcp-networks/{network_id}',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_dhcp-network',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_l3-router',
        check_str=base.ADMIN,
        description='Add a router to an L3 agent',
        operations=[
            {
                'method': 'POST',
                'path': '/agents/{agent_id}/l3-routers',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='create_l3-router',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_l3-routers',
        check_str=base.ADMIN,
        description='List routers on an L3 agent',
        operations=[
            {
                'method': 'GET',
                'path': '/agents/{agent_id}/l3-routers',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_l3-routers',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_l3-router',
        check_str=base.ADMIN,
        description='Remove a router from an L3 agent',
        operations=[
            {
                'method': 'DELETE',
                'path': '/agents/{agent_id}/l3-routers/{router_id}',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_l3-router',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_dhcp-agents',
        check_str=base.ADMIN,
        description='List DHCP agents hosting a network',
        operations=[
            {
                'method': 'GET',
                'path': '/networks/{network_id}/dhcp-agents',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_dhcp-agents',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_l3-agents',
        check_str=base.ADMIN,
        description='List L3 agents hosting a router',
        operations=[
            {
                'method': 'GET',
                'path': '/routers/{router_id}/l3-agents',
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_l3-agents',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



# Copyright 2024 Canonical Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import requests
import testtools
import webob

from oslo_config import cfg
from oslo_config import fixture as config_fixture

from neutron.agent.metadata import proxy_base
from neutron.common import utils
from neutron.conf.agent.metadata import config as meta_conf
from neutron.tests import base


class ConfFixture(config_fixture.Config):
    def setUp(self):
        super().setUp()
        meta_conf.register_meta_conf_opts(
            meta_conf.SHARED_OPTS, self.conf)
        meta_conf.register_meta_conf_opts(
            meta_conf.METADATA_PROXY_HANDLER_OPTS, self.conf)
        meta_conf.register_meta_conf_opts(
            meta_conf.UNIX_DOMAIN_METADATA_PROXY_OPTS, self.conf)
        self.config(auth_ca_cert=None,
                    nova_metadata_host='9.9.9.9',
                    nova_metadata_port=8775,
                    metadata_proxy_shared_secret='secret',
                    nova_metadata_protocol='http',
                    nova_metadata_insecure=True,
                    nova_client_cert='nova_cert',
                    nova_client_priv_key='nova_priv_key')


class FakeMetadataProxyHandler(proxy_base.MetadataProxyHandlerBase):
    def __init__(self, conf):
        super().__init__(conf)

    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        # This is an abstractmethod so must be defined
        return None, None


class TestMetadataProxyHandlerBase(base.BaseTestCase):
    fake_conf = cfg.CONF
    fake_conf_fixture = ConfFixture(fake_conf)

    def setUp(self):
        super().setUp()
        self.useFixture(self.fake_conf_fixture)
        self.handler = FakeMetadataProxyHandler(self.fake_conf)

    def _proxy_request_test_helper(self, response_code=200, method='GET'):
        hdrs = {'X-Forwarded-For': '8.8.8.8'}
        body = 'body'

        req = mock.Mock(path_info='/the_path', query_string='', headers=hdrs,
                        method=method, body=body)
        resp = mock.MagicMock(status_code=response_code)
        resp.status.__str__.side_effect = AttributeError
        resp.content = 'content'
        req.response = resp
        with mock.patch.object(utils, 'sign_instance_id') as sign:
            sign.return_value = 'signed'
            with mock.patch('requests.request') as mock_request:
                resp.headers = {'content-type': 'text/plain'}
                mock_request.return_value = resp
                retval = self.handler._proxy_request('the_id', 'tenant_id',
                                                     req)
                mock_request.assert_called_once_with(
                    method=method, url='http://9.9.9.9:8775/the_path',
                    headers={
                        'X-Forwarded-For': '8.8.8.8',
                        'X-Instance-ID-Signature': 'signed',
                        'X-Instance-ID': 'the_id',
                        'X-Tenant-ID': 'tenant_id'
                    },
                    data=body,
                    cert=(self.fake_conf.nova_client_cert,
                          self.fake_conf.nova_client_priv_key),
                    verify=False,
                    timeout=60)

                return retval

    def test_proxy_request_post(self):
        response = self._proxy_request_test_helper(method='POST')
        self.assertEqual('text/plain', response.content_type)
        self.assertEqual('content', response.body)

    def test_proxy_request_200(self):
        response = self._proxy_request_test_helper(200)
        self.assertEqual('text/plain', response.content_type)
        self.assertEqual('content', response.body)

    def test_proxy_request_400(self):
        self.assertIsInstance(self._proxy_request_test_helper(400),
                              webob.exc.HTTPBadRequest)

    def test_proxy_request_403(self):
        self.assertIsInstance(self._proxy_request_test_helper(403),
                              webob.exc.HTTPForbidden)

    def test_proxy_request_404(self):
        self.assertIsInstance(self._proxy_request_test_helper(404),
                              webob.exc.HTTPNotFound)

    def test_proxy_request_409(self):
        self.assertIsInstance(self._proxy_request_test_helper(409),
                              webob.exc.HTTPConflict)

    def test_proxy_request_500(self):
        self.assertIsInstance(self._proxy_request_test_helper(500),
                              webob.exc.HTTPInternalServerError)

    def test_proxy_request_502(self):
        self.assertIsInstance(self._proxy_request_test_helper(502),
                              webob.exc.HTTPBadGateway)

    def test_proxy_request_503(self):
        self.assertIsInstance(self._proxy_request_test_helper(503),
                              webob.exc.HTTPServiceUnavailable)

    def test_proxy_request_504(self):
        self.assertIsInstance(self._proxy_request_test_helper(504),
                              webob.exc.HTTPGatewayTimeout)

    def test_proxy_request_other_code(self):
        with testtools.ExpectedException(Exception):
            self._proxy_request_test_helper(302)

    def test_proxy_request_connection_error(self):
        req = mock.Mock(path_info='/the_path', query_string='', headers={},
                        method='GET', body='')
        with mock.patch('requests.request') as mock_request:
            mock_request.side_effect = requests.ConnectionError()
            retval = self.handler._proxy_request('the_id', 'tenant_id', req)
            self.assertIsInstance(retval, webob.exc.HTTPServiceUnavailable)


class FakeUnixDomainMetadataProxy(proxy_base.UnixDomainMetadataProxyBase):
    def run(self):
        # This is an abstractmethod so must be defined
        pass


class TestUnixDomainMetadataProxyBase(base.BaseTestCase):
    fake_conf = cfg.CONF
    fake_conf_fixture = ConfFixture(fake_conf)

    def setUp(self):
        super().setUp()
        self.useFixture(self.fake_conf_fixture)
        self.proxy = FakeUnixDomainMetadataProxy(self.fake_conf)

    def test__get_socket_mode_user(self):
        self.fake_conf_fixture.config(
            metadata_proxy_socket_mode=meta_conf.USER_MODE)
        mode = self.proxy._get_socket_mode()
        self.assertEqual(proxy_base.MODE_MAP[meta_conf.USER_MODE], mode)

    def test__get_socket_mode_deduce_user_root(self):
        self.fake_conf_fixture.config(
            metadata_proxy_socket_mode=meta_conf.DEDUCE_MODE,
            metadata_proxy_user='root')
        mode = self.proxy._get_socket_mode()
        self.assertEqual(proxy_base.MODE_MAP[meta_conf.USER_MODE], mode)

    @mock.patch.object(proxy_base.agent_utils, 'is_effective_group')
    @mock.patch.object(proxy_base.agent_utils, 'is_effective_user')
    def test__get_socket_mode_deduce_group(self, mock_ieu, mock_ieg):
        self.fake_conf_fixture.config(
            metadata_proxy_socket_mode=meta_conf.DEDUCE_MODE,
            metadata_proxy_user='fake',
            metadata_proxy_group='fake')
        # Always force non-effective user
        mock_ieu.return_value = False
        # group effective
        mock_ieg.return_value = True
        mode = self.proxy._get_socket_mode()
        self.assertEqual(proxy_base.MODE_MAP[meta_conf.GROUP_MODE], mode)
        # group non-effective
        mock_ieg.return_value = False
        mode = self.proxy._get_socket_mode()
        self.assertEqual(proxy_base.MODE_MAP[meta_conf.ALL_MODE], mode)



#
# Copyright (c) 2022, OVH SAS
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from oslo_config import cfg

from neutron.api import wsgi
from neutron.common import config
from neutron.common import profiler


def api_server():
    profiler.setup('neutron-server', cfg.CONF.host)
    app = config.load_paste_app('neutron')
    registry.publish(resources.PROCESS, events.BEFORE_SPAWN,
                     wsgi.WorkerService)
    registry.publish(resources.PROCESS, events.AFTER_INIT,
                     wsgi.WorkerService)
    return app



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import socketserver
import urllib

import jinja2
from neutron_lib.agent import topics
from neutron_lib import constants
from neutron_lib import context
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import encodeutils
import requests
import webob
from webob import exc as webob_exc

from neutron._i18n import _
from neutron.agent.common import base_agent_rpc
from neutron.agent.linux import utils as agent_utils
from neutron.agent.metadata import proxy_base
from neutron.agent import rpc as agent_rpc
from neutron.common import ipv6_utils
from neutron.common import utils as common_utils


LOG = logging.getLogger(__name__)

RESPONSE = jinja2.Template("""HTTP/1.1 {{ http_code }}
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: {{ len }}

<html>
 <head>
  <title>{{ title }}</title>
 </head>
 <body>
  <h1>{{ body_title }}</h1>
  {{ body }}<br /><br />


 </body>
</html>""")
RESPONSE_LENGHT = 40


class MetadataPluginAPI(base_agent_rpc.BasePluginApi):
    """Agent-side RPC for metadata agent-to-plugin interaction.

    This class implements the client side of an rpc interface used by the
    metadata service to make calls back into the Neutron plugin.  The server
    side is defined in
    neutron.api.rpc.handlers.metadata_rpc.MetadataRpcCallback.  For more
    information about changing rpc interfaces, see
    doc/source/contributor/internals/rpc_api.rst.

    API version history:
        1.0 - Initial version.
    """
    def __init__(self, topic):
        super().__init__(
            topic=topic,
            namespace=constants.RPC_NAMESPACE_METADATA,
            version='1.0')


class MetadataProxyHandlerBaseSocketServer(
        proxy_base.MetadataProxyHandlerBase):
    @staticmethod
    def _http_response(http_response, request):
        _res = webob.Response(
            body=http_response.content,
            status=http_response.status_code,
            content_type=http_response.headers['content-type'],
            charset=http_response.encoding)
        # NOTE(ralonsoh): there should be a better way to format the HTTP
        # response, adding the HTTP version to the ``webob.Response``
        # output string.
        out = request.http_version + ' ' + str(_res)
        if (int(_res.headers['content-length']) == 0 and
                _res.status_code == 200):
            # Add 2 extra \r\n to the result. HAProxy is also expecting
            # it even when the body is empty.
            out += '\r\n\r\n'
        return out.encode(http_response.encoding)

    def _proxy_request(self, instance_id, project_id, req):
        headers = {
            'X-Forwarded-For': req.headers.get('X-Forwarded-For'),
            'X-Instance-ID': instance_id,
            'X-Tenant-ID': project_id,
            'X-Instance-ID-Signature': common_utils.sign_instance_id(
                self.conf, instance_id)
        }

        nova_host_port = ipv6_utils.valid_ipv6_url(
            self.conf.nova_metadata_host,
            self.conf.nova_metadata_port)

        url = urllib.parse.urlunsplit((
            self.conf.nova_metadata_protocol,
            nova_host_port,
            req.path_info,
            req.query_string,
            ''))

        disable_ssl_certificate_validation = self.conf.nova_metadata_insecure
        if self.conf.auth_ca_cert and not disable_ssl_certificate_validation:
            verify_cert = self.conf.auth_ca_cert
        else:
            verify_cert = not disable_ssl_certificate_validation

        client_cert = None
        if self.conf.nova_client_cert and self.conf.nova_client_priv_key:
            client_cert = (self.conf.nova_client_cert,
                           self.conf.nova_client_priv_key)

        try:
            resp = requests.request(method=req.method, url=url,
                                    headers=headers,
                                    data=req.body,
                                    cert=client_cert,
                                    verify=verify_cert,
                                    timeout=60)
        except requests.ConnectionError:
            msg = _('The remote metadata server is temporarily unavailable. '
                    'Please try again later.')
            LOG.warning(msg)
            title = '503 Service Unavailable'
            length = RESPONSE_LENGHT + len(title) * 2 + len(msg)
            reponse = RESPONSE.render(http_code=title, title=title,
                                      body_title=title, body=title, len=length)
            return encodeutils.to_utf8(reponse)

        if resp.status_code == 200:
            return self._http_response(resp, req)
        if resp.status_code == 403:
            LOG.warning(
                'The remote metadata server responded with Forbidden. This '
                'response usually occurs when shared secrets do not match.'
            )
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code == 500:
            msg = _(
                'Remote metadata server experienced an internal server error.'
            )
            LOG.warning(msg)
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code in (400, 404, 409, 502, 503, 504):
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        raise Exception(_('Unexpected response code: %s') % resp.status_code)


class MetadataProxyHandler(MetadataProxyHandlerBaseSocketServer,
                           socketserver.StreamRequestHandler):
    NETWORK_ID_HEADER = 'X-Neutron-Network-ID'
    ROUTER_ID_HEADER = 'X-Neutron-Router-ID'
    _conf = None

    def __init__(self, request, client_address, server):
        self.plugin_rpc = MetadataPluginAPI(topics.PLUGIN)
        self.context = context.get_admin_context_without_session()
        super().__init__(self._conf, has_cache=False, request=request,
                         client_address=client_address, server=server)

    def handle(self):
        try:
            request = self.request.recv(4096)
            LOG.debug('Request: %s', request.decode('utf-8'))
            f_request = io.BytesIO(request)
            req = webob.Request.from_file(f_request)
            instance_id, project_id = self._get_instance_and_project_id(req)
            if instance_id:
                res = self._proxy_request(instance_id, project_id, req)
                self.wfile.write(res)
                return
            # TODO(ralonsoh): change this return to be a formatted Request
            # and added to self.wfile
            return webob_exc.HTTPNotFound()
        except Exception as exc:
            LOG.exception('Error while receiving data.')
            raise exc

    @staticmethod
    def _get_port_filters(router_id=None, ip_address=None,
                          networks=None, mac_address=None):
        filters = {}
        if router_id:
            filters['device_id'] = [router_id]
            filters['device_owner'] = constants.ROUTER_INTERFACE_OWNERS
        # We either get an IP assigned (and therefore known) by neutron
        # via X-Forwarded-For or that header contained a link-local
        # IPv6 address of which neutron only knows the MAC address encoded
        # in it. In the latter case the IPv6 address in X-Forwarded-For
        # is not a fixed ip of the port.
        if mac_address:
            filters['mac_address'] = [mac_address]
        elif ip_address:
            filters['fixed_ips'] = {'ip_address': [ip_address]}
        if networks:
            filters['network_id'] = networks

        return filters

    def _get_ports_from_server(self, router_id=None, ip_address=None,
                               networks=None, mac_address=None):
        """Get ports from server."""
        filters = self._get_port_filters(
            router_id, ip_address, networks, mac_address)
        return self.plugin_rpc.get_ports(self.context, filters)

    def _get_router_networks(self, router_id, skip_cache=False):
        """Find all networks connected to given router."""
        internal_ports = self._get_ports_from_server(router_id=router_id)
        return tuple(p['network_id'] for p in internal_ports)

    def _get_ports_for_remote_address(self, remote_address, networks,
                                      remote_mac=None,
                                      skip_cache=False):
        """Get list of ports that has given IP address and are part of
        given networks.

        :param remote_address: IP address to search for
        :param networks: List of networks in which the IP address will be
                         searched for
        :param remote_mac: Remote MAC to filter by, if given
        :param skip_cache: When to skip getting entry from cache

        """
        return self._get_ports_from_server(networks=networks,
                                           ip_address=remote_address,
                                           mac_address=remote_mac)

    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        if network_id:
            networks = (network_id,)
        elif router_id:
            networks = self._get_router_networks(router_id,
                                                 skip_cache=skip_cache)
        else:
            raise TypeError(_("Either one of parameter network_id or router_id"
                              " must be passed to get_port method."))

        ports = self._get_ports_for_remote_address(remote_address, networks,
                                                   remote_mac=remote_mac,
                                                   skip_cache=skip_cache)
        LOG.debug("Got ports for remote_address %(remote_address)s, "
                  "network_id %(network_id)s, remote_mac %(remote_mac)s, "
                  "router_id %(router_id)s"
                  "%(ports)s",
                  {"remote_address": remote_address,
                   "network_id": network_id,
                   "remote_mac": remote_mac,
                   "router_id": router_id,
                   "ports": ports})
        num_ports = len(ports)
        if num_ports == 1:
            return ports[0]['device_id'], ports[0]['tenant_id']
        if num_ports == 0:
            LOG.error("No port found in network %s with IP address %s",
                      network_id, remote_address)
        return None, None


class UnixDomainMetadataProxy(proxy_base.UnixDomainMetadataProxyBase):

    def __init__(self, conf):
        super().__init__(conf)
        agent_utils.ensure_directory_exists_without_file(
            cfg.CONF.metadata_proxy_socket)

    def _init_state_reporting(self):
        self.context = context.get_admin_context_without_session()
        self.failed_state_report = False
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        self.agent_state = {
            'binary': constants.AGENT_PROCESS_METADATA,
            'host': cfg.CONF.host,
            'topic': 'N/A',
            'configurations': {
                'metadata_proxy_socket': cfg.CONF.metadata_proxy_socket,
                'nova_metadata_host': cfg.CONF.nova_metadata_host,
                'nova_metadata_port': cfg.CONF.nova_metadata_port,
                'log_agent_heartbeats': cfg.CONF.AGENT.log_agent_heartbeats,
            },
            'start_flag': True,
            'agent_type': constants.AGENT_TYPE_METADATA}
        report_interval = cfg.CONF.AGENT.report_interval
        if report_interval:
            self.heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            self.heartbeat.start(interval=report_interval)

    def _report_state(self):
        try:
            self.state_rpc.report_state(
                self.context,
                self.agent_state,
                use_call=self.agent_state.get('start_flag'))
        except AttributeError:
            # This means the server does not support report_state
            LOG.warning('Neutron server does not support state report.'
                        ' State report for this agent will be disabled.')
            self.heartbeat.stop()
            return
        except Exception:
            self.failed_state_report = True
            LOG.exception("Failed reporting state!")
            return
        if self.failed_state_report:
            self.failed_state_report = False
            LOG.info('Successfully reported state after a previous failure.')
        self.agent_state.pop('start_flag', None)

    def run(self):
        file_socket = cfg.CONF.metadata_proxy_socket
        self._server = socketserver.ThreadingUnixStreamServer(
            file_socket, MetadataProxyHandler)
        MetadataProxyHandler._conf = self.conf
        self._server.serve_forever()



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from neutron.db import ovn_l3_hamode_db
from neutron.objects import router as router_obj
from neutron.tests.common import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit.db import test_l3_dvr_db


class FakeOVNL3Plugin(test_l3_dvr_db.FakeL3Plugin,
                      ovn_l3_hamode_db.OVN_L3_HA_db_mixin):
    pass


class OVN_L3_HA_db_mixinTestCase(test_plugin.NeutronDbPluginV2TestCase):

    def setUp(self, **kwargs):
        super().setUp(plugin='ml2', **kwargs)
        self.core_plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()
        self.mixin = FakeOVNL3Plugin()
        directory.add_plugin(plugin_constants.L3, self.mixin)

    def _create_router(self, router):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin._create_router_db(self.ctx, router, 'foo_tenant')

    def test_create_router(self):
        router_dict = {'name': 'foo_router', 'admin_state_up': True,
                       'distributed': False}
        router_db = self._create_router(router_dict)
        router = router_obj.Router.get_object(self.ctx, id=router_db.id)
        self.assertTrue(router.extra_attributes.ha)

    def test_create_no_ovn_router(self):
        router_dict = {'name': 'foo_router', 'admin_state_up': True,
                       'distributed': False, 'flavor_id': 'uuid'}
        router_db = self._create_router(router_dict)
        router = router_obj.Router.get_object(self.ctx, id=router_db.id)
        self.assertFalse(router.extra_attributes.ha)



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from oslo_config import fixture as fixture_config
from oslo_utils import uuidutils

from neutron.agent.ovn.agent import ovn_neutron_agent
from neutron.agent.ovn.agent import ovsdb as agent_ovsdb
from neutron.agent.ovn.metadata import agent as metadata_agent
from neutron.agent.ovn.metadata import server_socket
from neutron.common.ovn import constants as ovn_const
from neutron.common import utils as n_utils
from neutron.tests.common import net_helpers
from neutron.tests.functional import base


TEST_EXTENSION = 'testing'
METADATA_EXTENSION = 'metadata'
EXTENSION_NAMES = {TEST_EXTENSION: 'Fake OVN agent extension',
                   METADATA_EXTENSION: 'Metadata OVN agent extension',
                   }


class TestOVNNeutronAgentBase(base.TestOVNFunctionalBase):

    def setUp(self, extensions, **kwargs):
        super().setUp(**kwargs)
        self.host_name = 'host-' + uuidutils.generate_uuid()[:5]
        self.chassis_name = self.add_fake_chassis(self.host_name)
        self.extensions = extensions
        self.mock_chassis_name = mock.patch.object(
            agent_ovsdb, 'get_own_chassis_name',
            return_value=self.chassis_name).start()
        with mock.patch.object(metadata_agent.MetadataAgent,
                               '_get_own_chassis_name',
                               return_value=self.chassis_name):
            self.ovn_agent = self._start_ovn_neutron_agent()

    def _check_loaded_and_started_extensions(self, ovn_agent):
        for _ext in self.extensions:
            loaded_ext = ovn_agent[_ext]
            self.assertEqual(EXTENSION_NAMES.get(_ext), loaded_ext.name)
            self.assertTrue(loaded_ext.is_started)

    def _start_ovn_neutron_agent(self):
        conf = self.useFixture(fixture_config.Config()).conf
        conf.set_override('extensions', ','.join(self.extensions),
                          group='agent')
        ovn_nb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('nb')
        conf.set_override('ovn_nb_connection', [ovn_nb_db], group='ovn')
        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        conf.set_override('ovn_sb_connection', [ovn_sb_db], group='ovn')

        agt = ovn_neutron_agent.OVNNeutronAgent(conf)
        agt.test_ovs_idl = []
        agt.test_ovn_sb_idl = []
        agt.test_ovn_nb_idl = []
        # NOTE(ralonsoh): it is needed to ``UnixDomainMetadataProxy.wait``
        # method in eventlet environments in order not to block the execution.
        # Once eventlet is completely removed, this mock can be deleted.
        with mock.patch.object(ovn_neutron_agent.OVNNeutronAgent, 'wait'), \
                mock.patch.object(server_socket.UnixDomainMetadataProxy,
                                  'wait'):
            agt.start()
        self._check_loaded_and_started_extensions(agt)

        self.addCleanup(agt.ext_manager_api.ovs_idl.ovsdb_connection.stop)
        if agt.ext_manager_api.sb_idl:
            self.addCleanup(agt.ext_manager_api.sb_idl.ovsdb_connection.stop)
        if agt.ext_manager_api.nb_idl:
            self.addCleanup(agt.ext_manager_api.nb_idl.ovsdb_connection.stop)
        return agt


class TestOVNNeutronAgentFakeAgent(TestOVNNeutronAgentBase):

    def setUp(self, **kwargs):
        super().setUp(extensions=[TEST_EXTENSION], **kwargs)

    def test_ovs_and_ovs_events(self):
        # Test the OVS IDL is attending the provided events.
        bridge = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        exc = Exception('Bridge %s not added or not detected by '
                        'OVSInterfaceEvent' % bridge)
        n_utils.wait_until_true(
            lambda: bridge.br_name in self.ovn_agent.test_ovs_idl,
            timeout=10, exception=exc)

        # Test the OVN SB IDL is attending the provided events. The chassis is
        # created before the OVN SB IDL connection is created but the creation
        # event is received during the subscription.
        exc = Exception('Chassis %s not added or not detected by '
                        'OVNSBChassisEvent' % self.chassis_name)
        n_utils.wait_until_true(
            lambda: self.chassis_name in self.ovn_agent.test_ovn_sb_idl,
            timeout=10, exception=exc)

        # Test the OVN SN IDL is attending the provided events.
        lswitch_name = 'ovn-' + uuidutils.generate_uuid()
        self.nb_api.ls_add(lswitch_name).execute(check_error=True)
        exc = Exception('Logical Switch %s not added or not detected by ')
        n_utils.wait_until_true(
            lambda: lswitch_name in self.ovn_agent.test_ovn_nb_idl,
            timeout=10, exception=exc)


class TestOVNNeutronAgentMetadataExtension(TestOVNNeutronAgentBase):

    def setUp(self, **kwargs):
        super().setUp(extensions=[METADATA_EXTENSION], **kwargs)

    def test_check_metadata_started(self):
        # Check the metadata extension is registered.
        chassis_id = uuid.UUID(self.chassis_name)
        agent_id = uuid.uuid5(chassis_id, 'metadata_agent')
        ext_ids = {ovn_const.OVN_AGENT_METADATA_ID_KEY: str(agent_id)}
        ch_private = self.sb_api.lookup('Chassis_Private', self.chassis_name)
        self.assertEqual(ext_ids, ch_private.external_ids)

        # Check Unix proxy is running.
        metadata_extension = self.ovn_agent[METADATA_EXTENSION]
        self.assertIsNotNone(metadata_extension._proxy._server)



# Copyright (c) 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy as base_policy
import testscenarios

from neutron import policy
from neutron.tests.unit.conf.policies import test_base as base


class RbacAPITestCase(testscenarios.WithScenarios, base.PolicyBaseTestCase):

    scenarios = [
        ('target_tenant', {'_target_label': 'target_tenant'}),
        ('target_project', {'_target_label': 'target_project'})
    ]

    def setUp(self):
        super().setUp()
        self.target = {
            'project_id': self.project_id,
            self._target_label: 'other-project'}
        self.alt_target = {
            'project_id': self.alt_project_id,
            self._target_label: 'other-project'}
        self.wildcard_target = {
            'project_id': self.project_id,
            self._target_label: '*'}
        self.wildcard_alt_target = {
            'project_id': self.alt_project_id,
            self._target_label: '*'}


class SystemAdminTests(RbacAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.system_admin_ctx

    def test_create_rbac_policy(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy', self.alt_target)

    def test_create_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy:target_tenant',
            self.wildcard_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy:target_tenant',
            self.wildcard_alt_target)

    def test_create_rbac_policy_target_project(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy:target_project',
            self.wildcard_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_rbac_policy:target_project',
            self.wildcard_alt_target)

    def test_update_rbac_policy(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy', self.alt_target)

    def test_update_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy:target_tenant',
            self.wildcard_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy:target_tenant',
            self.wildcard_alt_target)

    def test_update_rbac_policy_target_project(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy:target_project',
            self.wildcard_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_rbac_policy:target_project',
            self.wildcard_alt_target)

    def test_get_rbac_policy(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_rbac_policy', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_rbac_policy', self.alt_target)


class SystemMemberTests(SystemAdminTests):

    def setUp(self):
        super().setUp()
        self.context = self.system_member_ctx


class SystemReaderTests(SystemMemberTests):

    def setUp(self):
        super().setUp()
        self.context = self.system_reader_ctx


class AdminTests(RbacAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.project_admin_ctx

    def test_create_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_rbac_policy', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_rbac_policy', self.alt_target))

    def test_create_rbac_policy_target_tenant(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_rbac_policy:target_tenant', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_rbac_policy:target_tenant', self.alt_target))

    def test_create_rbac_policy_target_project(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_rbac_policy:target_project', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_rbac_policy:target_project', self.alt_target))

    def test_update_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_rbac_policy', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_rbac_policy', self.alt_target))

    def test_update_rbac_policy_target_tenant(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_rbac_policy:target_tenant', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_rbac_policy:target_tenant', self.alt_target))

    def test_update_rbac_policy_target_project(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_rbac_policy:target_project', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_rbac_policy:target_project', self.alt_target))

    def test_get_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_rbac_policy', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_rbac_policy', self.alt_target))

    def test_delete_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_rbac_policy', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'delete_rbac_policy', self.alt_target))


class ProjectManagerTests(AdminTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_manager_ctx

    def test_create_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_rbac_policy', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy', self.alt_target)

    def test_create_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_tenant',
            self.wildcard_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_tenant',
            self.wildcard_alt_target)

    def test_create_rbac_policy_target_project(self):
        if 'target_tenant' in self.wildcard_target:
            self.skipTest('"create_rbac_policy:target_project" does not '
                          'support "target_tenant"')

        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_project',
            self.wildcard_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_project',
            self.wildcard_alt_target)

    def test_update_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_rbac_policy', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy', self.alt_target)

    def test_update_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_tenant',
            self.wildcard_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_tenant',
            self.wildcard_alt_target)

    def test_update_rbac_policy_target_project(self):
        if 'target_tenant' in self.wildcard_target:
            self.skipTest('"update_rbac_policy:target_project" does not '
                          'support "target_tenant"')

        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_project',
            self.wildcard_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_project',
            self.wildcard_alt_target)

    def test_get_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_rbac_policy', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_rbac_policy', self.alt_target)

    def test_delete_rbac_policy(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_rbac_policy', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_rbac_policy', self.alt_target)


class ProjectMemberTests(ProjectManagerTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_member_ctx


class ProjectReaderTests(ProjectMemberTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_reader_ctx

    def test_create_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy', self.alt_target)

    def test_update_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy', self.alt_target)

    def test_delete_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_rbac_policy', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_rbac_policy', self.alt_target)


class ServiceRoleTests(RbacAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.service_ctx

    def test_create_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy', self.target)

    def test_create_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_tenant',
            self.wildcard_target)

    def test_create_rbac_policy_target_project(self):
        if 'target_tenant' in self.wildcard_target:
            self.skipTest('"create_rbac_policy:target_project" does not '
                          'support "target_tenant"')

        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_rbac_policy:target_project',
            self.wildcard_target)

    def test_update_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy', self.target)

    def test_update_rbac_policy_target_tenant(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_tenant',
            self.wildcard_target)

    def test_update_rbac_policy_target_project(self):
        if 'target_tenant' in self.wildcard_target:
            self.skipTest('"update_rbac_policy:target_project" does not '
                          'support "target_tenant"')

        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_rbac_policy:target_project',
            self.wildcard_target)

    def test_get_rbac_policy(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_rbac_policy', self.target)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_config import cfg

from neutron._i18n import _
from neutron.common import _constants as common_const


ml2_opts = [
    cfg.ListOpt('type_drivers',
                default=[constants.TYPE_LOCAL, constants.TYPE_FLAT,
                         constants.TYPE_VLAN, constants.TYPE_GRE,
                         constants.TYPE_VXLAN, constants.TYPE_GENEVE],
                help=_("List of network type driver entrypoints to be loaded "
                       "from the neutron.ml2.type_drivers namespace.")),
    cfg.ListOpt('tenant_network_types',
                default=[constants.TYPE_LOCAL],
                help=_("Ordered list of network_types to allocate as tenant "
                       "networks. The default value 'local' is useful for "
                       "single-box testing but provides no connectivity "
                       "between hosts.")),
    cfg.ListOpt('mechanism_drivers',
                default=[],
                help=_("An ordered list of networking mechanism driver "
                       "entrypoints to be loaded from the "
                       "neutron.ml2.mechanism_drivers namespace.")),
    cfg.ListOpt('extension_drivers',
                default=[],
                help=_("An ordered list of extension driver "
                       "entrypoints to be loaded from the "
                       "neutron.ml2.extension_drivers namespace. "
                       "For example: extension_drivers = port_security,qos")),
    cfg.IntOpt('path_mtu', default=0,
               help=_('Maximum size of an IP packet (MTU) that can traverse '
                      'the underlying physical network infrastructure without '
                      'fragmentation when using an overlay/tunnel protocol. '
                      'This option allows specifying a physical network MTU '
                      'value that differs from the default global_physnet_mtu '
                      'value.')),
    cfg.DictOpt('physical_network_mtus',
                default={},
                help=_("Mappings of physical networks to MTU values. "
                       "The format of the mapping is <physnet>:<mtu val>. "
                       "This mapping allows specifying a physical network MTU "
                       "value that differs from "
                       "the default global_physnet_mtu value.")),
    cfg.StrOpt('external_network_type',
               help=_("Default network type for external networks when no "
                      "provider attributes are specified. By default it is "
                      "None, which means that if provider attributes are not "
                      "specified while creating external networks then they "
                      "will have the same type as tenant networks. Allowed "
                      "values for external_network_type config option depend "
                      "on the network type values configured in type_drivers "
                      "config option.")),
    cfg.IntOpt('overlay_ip_version',
               default=constants.IP_VERSION_4,
               choices=[
                   (constants.IP_VERSION_4, 'IPv4'),
                   (constants.IP_VERSION_6, 'IPv6')
               ],
               help=_("IP version of all overlay (tunnel) network "
                      "endpoints.")),
    cfg.StrOpt('tunnelled_network_rp_name',
               default=common_const.RP_TUNNELLED,
               help=_("Resource provider name for the host with tunnelled "
                      "networks. This resource provider represents the "
                      "available bandwidth for all tunnelled networks in a "
                      "compute node. NOTE: this parameter is used both by the "
                      "Neutron server and the mechanism driver agents; it is "
                      "recommended not to change it once any resource "
                      "provider register has been created.")),
]


def register_ml2_plugin_opts(cfg=cfg.CONF):
    cfg.register_opts(ml2_opts, "ml2")



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.db import resource_extend
from neutron_lib.services.qos import constants as qos_consts


@resource_extend.has_resource_extenders
class FloatingQoSDbMixin:
    """Mixin class to enable floating IP's QoS extra attributes."""

    @staticmethod
    @resource_extend.extends([l3_apidef.FLOATINGIPS])
    def _extend_extra_fip_dict(fip_res, fip_db):
        qos_id = (fip_db.qos_policy_binding.policy_id if
                  fip_db.qos_policy_binding else None)
        fip_res[qos_consts.QOS_POLICY_ID] = qos_id
        qos_id = (fip_db.qos_network_policy_binding.policy_id if
                  fip_db.qos_network_policy_binding else None)
        fip_res[qos_consts.QOS_NETWORK_POLICY_ID] = qos_id
        return fip_res



# Copyright (C) 2014,2015 VA Linux Systems Japan K.K.
# Copyright (C) 2014,2015 YAMAMOTO Takashi <yamamoto at valinux co jp>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.plugins.ml2 import ovs_constants as ovs_const

from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent.openflow.native \
    import ovs_bridge_test_base


call = mock.call  # short hand


class OVSPhysicalBridgeTest(ovs_bridge_test_base.OVSBridgeTestBase,
                            ovs_bridge_test_base.OVSDVRProcessTestMixin):
    dvr_process_table_id = ovs_const.DVR_PROCESS_PHYSICAL
    dvr_process_next_table_id = ovs_const.LOCAL_VLAN_TRANSLATION

    def setUp(self):
        conn_patcher = mock.patch(
            'neutron.agent.ovsdb.impl_idl._connection')
        conn_patcher.start()
        super().setUp()
        self.addCleanup(conn_patcher.stop)
        self.setup_bridge_mock('br-phys', self.br_phys_cls)
        self.stamp = self.br.default_cookie

    def test_setup_default_table(self):
        self.br.setup_default_table()
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(
                ofpp.OFPFlowMod(
                    dp,
                    cookie=self.stamp,
                    instructions=[
                        ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                            ofpp.OFPActionOutput(ofp.OFPP_NORMAL, 0),
                        ]),
                    ],
                    match=ofpp.OFPMatch(),
                    priority=0,
                    table_id=0),
                active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_provision_local_vlan(self):
        port = 999
        lvid = 888
        segmentation_id = 777
        distributed = False
        self.br.provision_local_vlan(port=port, lvid=lvid,
                                     segmentation_id=segmentation_id,
                                     distributed=distributed)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(
                ofpp.OFPFlowMod(
                    dp,
                    cookie=self.stamp,
                    instructions=[
                        ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                            ofpp.OFPActionSetField(
                                vlan_vid=segmentation_id | ofp.OFPVID_PRESENT),
                            ofpp.OFPActionOutput(ofp.OFPP_NORMAL, 0),
                        ]),
                    ],
                    match=ofpp.OFPMatch(
                        in_port=port,
                        vlan_vid=lvid | ofp.OFPVID_PRESENT),
                    priority=4,
                    table_id=0),
                active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_provision_local_vlan_novlan(self):
        port = 999
        lvid = 888
        segmentation_id = None
        distributed = False
        self.br.provision_local_vlan(port=port, lvid=lvid,
                                     segmentation_id=segmentation_id,
                                     distributed=distributed)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(
                ofpp.OFPFlowMod(
                    dp,
                    cookie=self.stamp,
                    instructions=[
                        ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                            ofpp.OFPActionPopVlan(),
                            ofpp.OFPActionOutput(ofp.OFPP_NORMAL, 0),
                        ]),
                    ],
                    match=ofpp.OFPMatch(
                        in_port=port,
                        vlan_vid=lvid | ofp.OFPVID_PRESENT),
                    priority=4,
                    table_id=0),
                active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_reclaim_local_vlan(self):
        port = 999
        lvid = 888
        self.br.reclaim_local_vlan(port=port, lvid=lvid)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(
                match=ofpp.OFPMatch(
                    in_port=port,
                    vlan_vid=lvid | ofp.OFPVID_PRESENT)),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_add_dvr_mac_physical(self):
        mac = '00:02:b3:13:fe:3d'
        port = 8888
        self.br.add_dvr_mac_physical(mac=mac, port=port)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(
                ofpp.OFPFlowMod(
                    dp,
                    cookie=self.stamp,
                    instructions=[
                        ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                            ofpp.OFPActionOutput(port, 0),
                        ]),
                    ],
                    match=ofpp.OFPMatch(eth_src=mac),
                    priority=2,
                    table_id=3),
                active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_remove_dvr_mac_vlan(self):
        mac = '00:02:b3:13:fe:3d'
        self.br.remove_dvr_mac_vlan(mac=mac)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(eth_src=mac, table_id=3),
        ]
        self.assertEqual(expected, self.mock.mock_calls)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.linux import utils as agent_utils
from neutron.agent.metadata import proxy_base
from neutron.agent.ovn.metadata import ovsdb
from neutron.common.ovn import constants as ovn_const


LOG = logging.getLogger(__name__)


class MetadataProxyHandler(proxy_base.MetadataProxyHandlerBase):
    NETWORK_ID_HEADER = 'X-OVN-Network-ID'
    ROUTER_ID_HEADER = ''

    def __init__(self, conf, chassis, sb_idl):
        super().__init__(conf)
        self.chassis = chassis
        self._sb_idl = sb_idl
        self._post_fork_event = threading.Event()
        self.subscribe()

    @property
    def sb_idl(self):
        if not self._sb_idl:
            self._post_fork_event.wait()

        return self._sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self._sb_idl = val

    def subscribe(self):
        registry.subscribe(self.post_fork_initialize,
                           resources.PROCESS,
                           events.AFTER_INIT)

    def post_fork_initialize(self, resource, event, trigger, payload=None):
        # We need to open a connection to OVN SouthBound database for
        # each worker so that we can process the metadata requests.
        self._post_fork_event.clear()
        self.sb_idl = ovsdb.MetadataAgentOvnSbIdl(
            tables=('Port_Binding', 'Datapath_Binding', 'Chassis'),
            chassis=self.chassis).start()

        # Now IDL connections can be safely used.
        self._post_fork_event.set()

    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        ports = self.sb_idl.get_network_port_bindings_by_ip(network_id,
                                                            remote_address,
                                                            mac=remote_mac)
        num_ports = len(ports)
        if num_ports == 1:
            external_ids = ports[0].external_ids
            return (external_ids[ovn_const.OVN_DEVID_EXT_ID_KEY],
                    external_ids[ovn_const.OVN_PROJID_EXT_ID_KEY])
        if num_ports == 0:
            LOG.error("No port found in network %s with IP address %s",
                      network_id, remote_address)
        elif num_ports > 1:
            port_uuids = ', '.join([str(port.uuid) for port in ports])
            LOG.error("More than one port found in network %s with IP address "
                      "%s. Please run the neutron-ovn-db-sync-util script as "
                      "there seems to be inconsistent data between Neutron "
                      "and OVN databases. OVN Port uuids: %s", network_id,
                      remote_address, port_uuids)
        return None, None


class UnixDomainMetadataProxy(proxy_base.UnixDomainMetadataProxyBase):

    def __init__(self, conf, chassis, sb_idl=None):
        super().__init__(conf)
        self.chassis = chassis
        self.sb_idl = sb_idl
        agent_utils.ensure_directory_exists_without_file(
            cfg.CONF.metadata_proxy_socket)

    def run(self):
        self.server = agent_utils.UnixDomainWSGIServer(
            'neutron-ovn-metadata-agent')
        # Set the default metadata_workers if not yet set in the config file
        md_workers = self.conf.metadata_workers
        md_workers = 0 if md_workers is None else md_workers
        sb_idl = self.sb_idl if md_workers == 0 else None
        self.server.start(MetadataProxyHandler(self.conf, self.chassis,
                                               sb_idl),
                          self.conf.metadata_proxy_socket,
                          workers=md_workers,
                          backlog=self.conf.metadata_backlog,
                          mode=self._get_socket_mode())

    def wait(self):
        self.server.wait()



# Copyright (c) 2016 Mellanox Technologies, Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutron_lib.agent import l2_extension
from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants
from neutron_lib.utils import helpers
from oslo_config import cfg
from oslo_log import log as logging
from pyroute2.netlink import exceptions as netlink_exceptions

from neutron.agent.linux import bridge_lib
from neutron.conf.agent import l2_ext_fdb_population

l2_ext_fdb_population.register_fdb_population_opts()

LOG = logging.getLogger(__name__)


class FdbPopulationAgentExtension(
        l2_extension.L2AgentExtension):
    """The FDB population is an agent extension to OVS
    whose objective is to update the FDB table for existing instance
    using normal port, thus enabling communication between SR-IOV instances
    and normal instances.
    Additional information describing the problem can be found here:
    http://events.linuxfoundation.org/sites/events/files/slides/LinuxConJapan2014_makita_0.pdf
    """

    # FDB updates are triggered for ports with a certain device_owner only:
    # - device owner "compute": updates the FDB with normal port instances,
    #       required in order to enable communication between
    #       SR-IOV direct port instances and normal port instance.
    # - device owner "router_interface": updates the FDB with OVS/LB ports,
    #       required in order to enable communication for SR-IOV instances
    #       with floating ip that are located with the network node.
    # - device owner "DHCP": updates the FDB with the dhcp server.
    #       When the lease expires a unicast renew message is sent
    #       to the dhcp server. In case the FDB is not updated
    #       the message will be sent to the wire, causing the message
    #       to get lost in case the sender uses direct port and is
    #       located on the same hypervisor as the network node.
    PERMITTED_DEVICE_OWNERS = {constants.DEVICE_OWNER_COMPUTE_PREFIX,
                               constants.DEVICE_OWNER_ROUTER_INTF,
                               constants.DEVICE_OWNER_DHCP}

    class FdbTableTracker:
        """FDB table tracker is a helper class
        intended to keep track of the existing FDB rules.
        """
        def __init__(self, devices):
            self.device_to_macs = {}
            self.portid_to_mac = {}
            # update macs already in the physical interface's FDB table
            for device in devices:
                try:
                    rules = bridge_lib.FdbInterface.show(dev=device)
                except (OSError, netlink_exceptions.NetlinkError) as e:
                    LOG.warning(
                        'Unable to find FDB Interface %(device)s. '
                        'Exception: %(e)s', {'device': device, 'e': e})
                    continue
                self.device_to_macs[device] = [rule['mac'] for rule in
                                               rules[device]]

        def update_port(self, device, port_id, mac):
            # check if device is updated
            if self.device_to_macs.get(device) == mac:
                return
            # delete invalid port_id's mac from the FDB,
            # in case the port was updated to another mac
            self.delete_port([device], port_id)
            # update port id
            self.portid_to_mac[port_id] = mac
            # check if rule for mac already exists
            if mac in self.device_to_macs[device]:
                return
            if not bridge_lib.FdbInterface.add(mac, device):
                LOG.warning('Unable to add mac %(mac)s to FDB Interface '
                            '%(device)s.', {'mac': mac, 'device': device})
                return
            self.device_to_macs[device].append(mac)

        def delete_port(self, devices, port_id):
            mac = self.portid_to_mac.get(port_id)
            if mac is None:
                LOG.warning('Port Id %(port_id)s does not have a rule for '
                            'devices %(devices)s in FDB table',
                            {'port_id': port_id, 'devices': devices})
                return
            for device in devices:
                if mac in self.device_to_macs[device]:
                    if not bridge_lib.FdbInterface.delete(mac, device):
                        LOG.warning('Unable to delete mac %(mac)s from FDB '
                                    'Interface %(device)s.',
                                    {'mac': mac, 'device': device})
                        return
                    self.device_to_macs[device].remove(mac)
                    del self.portid_to_mac[port_id]

    # class FdbPopulationAgentExtension implementation:
    def initialize(self, connection, driver_type):
        """Perform FDB Agent Extension initialization."""
        if driver_type != ovs_constants.EXTENSION_DRIVER_TYPE:
            LOG.error('FDB extension is only supported for OVS agent, '
                      f'currently uses {driver_type}')
            sys.exit(1)

        self.device_mappings = helpers.parse_mappings(
            cfg.CONF.FDB.shared_physical_device_mappings, unique_keys=False)
        devices = self._get_devices()
        if not devices:
            LOG.error('Invalid configuration provided for FDB extension: '
                      'no physical devices')
            sys.exit(1)
        self.fdb_tracker = self.FdbTableTracker(devices)

    def handle_port(self, context, details):
        """Handle agent FDB population extension for port."""
        device_owner = details['device_owner']
        if self._is_valid_device_owner(device_owner):
            mac = details['mac_address']
            port_id = details['port_id']
            physnet = details.get('physical_network')
            if physnet and physnet in self.device_mappings:
                for device in self.device_mappings[physnet]:
                    self.fdb_tracker.update_port(device, port_id, mac)

    def delete_port(self, context, details):
        """Delete port from FDB population extension."""
        port_id = details['port_id']
        devices = self._get_devices()
        self.fdb_tracker.delete_port(devices, port_id)

    def _get_devices(self):
        def _flatten_list(li):
            return [item for sublist in li for item in sublist]

        return _flatten_list(self.device_mappings.values())

    def _is_valid_device_owner(self, device_owner):
        for permitted_device_owner in self.PERMITTED_DEVICE_OWNERS:
            if device_owner.startswith(permitted_device_owner):
                return True
        return False



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import grp
import os
import pwd
import signal

from neutron_lib import exceptions
from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent.linux import external_process
from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as linux_utils
from neutron.common import metadata as comm_meta
from neutron.common import utils as common_utils


LOG = logging.getLogger(__name__)

SIGTERM_TIMEOUT = 5

METADATA_SERVICE_NAME = 'metadata-proxy'
HAPROXY_SERVICE = 'haproxy'

_UNLIMITED_CONFIG_TEMPLATE = """
listen listener
    bind %(host)s:%(port)s
    %(bind_v6_line)s
    server metadata %(unix_socket_path)s
"""


class HaproxyConfiguratorBase(metaclass=abc.ABCMeta):
    PROXY_CONFIG_DIR: str
    HEADER_CONFIG_TEMPLATE: str

    def __init__(self, network_id, router_id, unix_socket_path, host, port,
                 user, group, state_path, pid_file, rate_limiting_config,
                 host_v6=None, bind_interface=None):
        self.network_id = network_id
        self.router_id = router_id
        if network_id is None and router_id is None:
            raise exceptions.NetworkIdOrRouterIdRequiredError()

        self.host = host
        self.host_v6 = host_v6
        self.bind_interface = bind_interface
        self.port = port
        self.user = user
        self.group = group
        self.state_path = state_path
        self.unix_socket_path = unix_socket_path
        self.pidfile = pid_file
        self.rate_limiting_config = rate_limiting_config
        self.log_level = (
            'debug' if logging.is_debug_enabled(cfg.CONF) else 'info')
        # log-tag will cause entries to have the string pre-pended, so use
        # the uuid haproxy will be started with.  Additionally, if it
        # starts with "haproxy" then things will get logged to
        # /var/log/haproxy.log on Debian distros, instead of to syslog.
        uuid = network_id or router_id
        self.log_tag = f"haproxy-{METADATA_SERVICE_NAME}-{uuid}"
        self._haproxy_cfg = ''
        self._resource_id = None
        self._create_config()

    @property
    def haproxy_cfg(self) -> str:
        return self._haproxy_cfg

    @property
    def resource_id(self) -> str:
        return self._resource_id

    def _create_config(self) -> None:
        """Create the configuration for haproxy, stored locally

        This method creates a string with the HAProxy configuration, stored in
        ``self._haproxy_cfg``. It also stores the resource ID (network, router)
        in ``self._resource_id``.

        This method must be called once in the init method.
        """
        # Need to convert uid/gid into username/group
        try:
            username = pwd.getpwuid(int(self.user)).pw_name
        except (ValueError, KeyError):
            try:
                username = pwd.getpwnam(self.user).pw_name
            except KeyError:
                raise comm_meta.InvalidUserOrGroupException(
                    _("Invalid user/uid: '%s'") % self.user)

        try:
            groupname = grp.getgrgid(int(self.group)).gr_name
        except (ValueError, KeyError):
            try:
                groupname = grp.getgrnam(self.group).gr_name
            except KeyError:
                raise comm_meta.InvalidUserOrGroupException(
                    _("Invalid group/gid: '%s'") % self.group)

        cfg_info = {
            'host': self.host,
            'port': self.port,
            'unix_socket_path': self.unix_socket_path,
            'user': username,
            'group': groupname,
            'pidfile': self.pidfile,
            'log_level': self.log_level,
            'log_tag': self.log_tag,
            'bind_v6_line': '',
        }
        if self.host_v6 and self.bind_interface:
            cfg_info['bind_v6_line'] = (
                'bind {}:{} interface {}'.format(
                    self.host_v6, self.port, self.bind_interface)
            )
        # If using the network ID, delete any spurious router ID that might
        # have been in the request, same for network ID when using router ID.
        # This is to prevent someone from spoofing a metadata request using
        # the proxy via an external network. See LP #1865036 for more info.
        # This only applies to the non-OVN driver.
        if self.network_id:
            cfg_info['res_type'] = 'Network'
            cfg_info['res_id'] = self.network_id
            cfg_info['res_type_del'] = 'Router'
        else:
            cfg_info['res_type'] = 'Router'
            cfg_info['res_id'] = self.router_id
            cfg_info['res_type_del'] = 'Network'
        self._resource_id = cfg_info['res_id']
        self._haproxy_cfg = comm_meta.get_haproxy_config(
            cfg_info, self.rate_limiting_config,
            self.HEADER_CONFIG_TEMPLATE, _UNLIMITED_CONFIG_TEMPLATE)

    def create_config_file(self):
        """Read the configuration stored and write the configuration file"""
        LOG.debug("haproxy_cfg = %s", self.haproxy_cfg)
        cfg_dir = self.get_config_path(self.state_path)
        # uuid has to be included somewhere in the command line so that it can
        # be tracked by process_monitor.
        self.cfg_path = os.path.join(cfg_dir, "%s.conf" % self.resource_id)
        if not os.path.exists(cfg_dir):
            os.makedirs(cfg_dir)
        with open(self.cfg_path, "w") as cfg_file:
            cfg_file.write(self.haproxy_cfg)

    @classmethod
    def get_config_path(cls, state_path):
        return os.path.join(state_path or cfg.CONF.state_path,
                            cls.PROXY_CONFIG_DIR)

    def read_config_file(self) -> str:
        """Return a string with the content of the configuration file"""
        cfg_path = os.path.join(self.get_config_path(self.state_path),
                                '%s.conf' % self.resource_id)
        return linux_utils.read_if_exists(str(cfg_path), run_as_root=True)

    def is_config_file_obsolete(self) -> bool:
        """Compare the instance config and the config file content

        Returns False if both configurations match. This check skips the
        "pidfile" line because that is provided just before the process is
        started.
        """
        def trim_config(haproxy_cfg: str) -> list[str]:
            return [line for line in haproxy_cfg.split('\n')
                    if not line.lstrip().startswith('pidfile')]

        file_config = trim_config(self.read_config_file())
        current_config = trim_config(self.haproxy_cfg)
        return file_config != current_config

    @classmethod
    def cleanup_config_file(cls, uuid, state_path):
        """Delete config file created when metadata proxy was spawned."""
        # Delete config file if it exists
        cfg_path = os.path.join(
            cls.get_config_path(state_path),
            "%s.conf" % uuid)
        linux_utils.delete_if_exists(cfg_path, run_as_root=True)


class MetadataDriverBase(metaclass=abc.ABCMeta):
    monitors = {}

    @staticmethod
    @abc.abstractmethod
    def haproxy_configurator():
        """Returns the HaproxyConfigurator for the class."""
        pass

    @classmethod
    def _get_metadata_proxy_user_group(cls, conf):
        user = conf.metadata_proxy_user or str(os.geteuid())
        group = conf.metadata_proxy_group or str(os.getegid())

        return user, group

    @classmethod
    def _get_haproxy_configurator(cls, bind_address, port, conf,
                                  network_id=None, router_id=None,
                                  bind_address_v6=None,
                                  bind_interface=None,
                                  pid_file=''):
        metadata_proxy_socket = conf.metadata_proxy_socket
        user, group = cls._get_metadata_proxy_user_group(conf)
        configurator = cls.haproxy_configurator()
        return configurator(network_id,
                            router_id,
                            metadata_proxy_socket,
                            bind_address,
                            port,
                            user,
                            group,
                            conf.state_path,
                            pid_file,
                            conf.metadata_rate_limiting,
                            bind_address_v6,
                            bind_interface)

    @classmethod
    def _get_metadata_proxy_callback(cls, bind_address, port, conf,
                                     network_id=None, router_id=None,
                                     bind_address_v6=None,
                                     bind_interface=None):
        def callback(pid_file):
            haproxy = cls._get_haproxy_configurator(
                bind_address, port, conf, network_id, router_id,
                bind_address_v6, bind_interface, pid_file)
            haproxy.create_config_file()
            return [HAPROXY_SERVICE, '-f', haproxy.cfg_path]

        return callback

    @classmethod
    def spawn_monitored_metadata_proxy(cls, monitor, ns_name, port, conf,
                                       bind_address="0.0.0.0", network_id=None,
                                       router_id=None, bind_address_v6=None,
                                       bind_interface=None):
        if bind_interface is not None and bind_address_v6 is not None:
            # HAProxy cannot bind() until IPv6 Duplicate Address Detection
            # completes. We must wait until the address leaves its 'tentative'
            # state.
            try:
                ip_lib.IpAddrCommand(
                    parent=ip_lib.IPDevice(name=bind_interface,
                                           namespace=ns_name)
                ).wait_until_address_ready(address=bind_address_v6)
            except ip_lib.DADFailed as exc:
                # This failure means that another DHCP agent has already
                # configured this metadata address, so all requests will
                # be via that single agent.
                LOG.info('DAD failed for address %(address)s on interface '
                         '%(interface)s in namespace %(namespace)s on network '
                         '%(network)s, deleting it. Exception: %(exception)s',
                         {'address': bind_address_v6,
                          'interface': bind_interface,
                          'namespace': ns_name,
                          'network': network_id,
                          'exception': str(exc)})
                try:
                    ip_lib.delete_ip_address(bind_address_v6, bind_interface,
                                             namespace=ns_name)
                except Exception as exc:
                    # do not re-raise a delete failure, just log
                    LOG.info('Address deletion failure: %s', str(exc))

                # Do not use the address or interface when DAD fails
                bind_address_v6 = bind_interface = None

        # If the HAProxy running instance configuration is different from
        # the one passed in this call, the HAProxy is stopped. The new
        # configuration will be written to the disk and a new instance
        # started.
        haproxy_cfg = cls._get_haproxy_configurator(
            bind_address, port, conf, network_id=network_id,
            router_id=router_id, bind_address_v6=bind_address_v6,
            bind_interface=bind_interface)
        if haproxy_cfg.is_config_file_obsolete():
            cls.destroy_monitored_metadata_proxy(
                monitor, haproxy_cfg.resource_id, conf, ns_name)

        uuid = network_id or router_id
        callback = cls._get_metadata_proxy_callback(
            bind_address, port, conf,
            network_id=network_id, router_id=router_id,
            bind_address_v6=bind_address_v6, bind_interface=bind_interface)
        pm = cls._get_metadata_proxy_process_manager(uuid, conf,
                                                     ns_name=ns_name,
                                                     callback=callback)
        try:
            pm.enable(ensure_active=True)
        except exceptions.ProcessExecutionError as exec_err:
            LOG.error("Encountered process execution error %(err)s while "
                      "starting process in namespace %(ns)s",
                      {"err": exec_err, "ns": ns_name})
            return
        monitor.register(uuid, METADATA_SERVICE_NAME, pm)

        cls.monitors[uuid] = pm

    @classmethod
    def destroy_monitored_metadata_proxy(cls, monitor, uuid, conf, ns_name):
        monitor.unregister(uuid, METADATA_SERVICE_NAME)
        pm = cls._get_metadata_proxy_process_manager(uuid, conf,
                                                     ns_name=ns_name)
        pm.disable(sig=str(int(signal.SIGTERM)))
        try:
            common_utils.wait_until_true(lambda: not pm.active,
                                         timeout=SIGTERM_TIMEOUT)
        except common_utils.WaitTimeout:
            LOG.warning('Metadata process %s did not finish after SIGTERM '
                        'signal in %s seconds, sending SIGKILL signal',
                        pm.pid, SIGTERM_TIMEOUT)
            pm.disable(sig=str(int(signal.SIGKILL)))

        # Delete metadata proxy config.
        configurator = cls.haproxy_configurator()
        configurator.cleanup_config_file(uuid, cfg.CONF.state_path)

        cls.monitors.pop(uuid, None)

    @classmethod
    def _get_metadata_proxy_process_manager(cls, router_id, conf, ns_name=None,
                                            callback=None):
        return external_process.ProcessManager(
            conf=conf,
            uuid=router_id,
            namespace=ns_name,
            service=HAPROXY_SERVICE,
            default_cmd_callback=callback)



# Copyright (c) 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import threading

from neutron_lib.agent import extension
from neutron_lib import exceptions
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent import agent_extensions_manager as agent_ext_mgr


LOG = logging.getLogger(__name__)
OVN_AGENT_EXT_MANAGER_NAMESPACE = 'neutron.agent.ovn.extensions'


class ConfigException(exceptions.NeutronException):
    """Misconfiguration of the OVN Neutron Agent"""
    message = _('Error configuring the OVN Neutron Agent: %(description)s.')


class OVNAgentExtensionManager(agent_ext_mgr.AgentExtensionsManager):

    def __init__(self, conf):
        super().__init__(conf, OVN_AGENT_EXT_MANAGER_NAMESPACE)
        for ext in self:
            if not isinstance(ext.obj, OVNAgentExtension):
                desc = ('Extension %s class is not inheriting from '
                        '"OVNAgentExtension"')
                raise ConfigException(description=desc)

    def start(self):
        """Start the extensions, once the OVN agent has been initialized."""
        for ext in self:
            ext.obj.start()
            LOG.info('Extension manager: %s started', ext.obj.name)


class OVNAgentExtension(extension.AgentExtension, metaclass=abc.ABCMeta):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.agent_api = None
        self._is_started = False

    @property
    @abc.abstractmethod
    def name(self):
        pass

    def initialize(self, *args):
        """Initialize agent extension."""
        pass

    def consume_api(self, agent_api):
        """Configure the Agent API.

        Allows an extension to gain access to resources internal to the
        neutron agent and otherwise unavailable to the extension.
        """
        self.agent_api = agent_api

    def start(self):
        """Start the extension, once the OVN agent has been initialized.

        This method executes any action needed after the initialization of the
        OVN agent and the extension manager API. It is executed at the end of
        the OVN agent ``start`` method.
        """
        self._is_started = True

    @property
    def is_started(self):
        return self._is_started

    @property
    @abc.abstractmethod
    def ovs_idl_events(self):
        pass

    @property
    @abc.abstractmethod
    def nb_idl_tables(self):
        pass

    @property
    @abc.abstractmethod
    def nb_idl_events(self):
        pass

    @property
    @abc.abstractmethod
    def sb_idl_tables(self):
        pass

    @property
    @abc.abstractmethod
    def sb_idl_events(self):
        pass


class OVNAgentExtensionAPI:
    """Implements the OVN Neutron Agent API"""

    def __init__(self):
        self._nb_idl = None
        self._sb_idl = None
        self._has_chassis_private = None
        self._ovs_idl = None
        self.sb_post_fork_event = threading.Event()
        self.sb_post_fork_event.clear()
        self.nb_post_fork_event = threading.Event()
        self.nb_post_fork_event.clear()

    @property
    def ovs_idl(self):
        return self._ovs_idl

    @ovs_idl.setter
    def ovs_idl(self, val):
        self._ovs_idl = val

    @property
    def nb_idl(self):
        if not self._nb_idl:
            self.nb_post_fork_event.wait()
        return self._nb_idl

    @nb_idl.setter
    def nb_idl(self, val):
        self.nb_post_fork_event.set()
        self._nb_idl = val

    @property
    def sb_idl(self):
        if not self._sb_idl:
            self.sb_post_fork_event.wait()
        return self._sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self.sb_post_fork_event.set()
        self._sb_idl = val
